/*
 * Tue Apr 13 10:55:44 CEST 2004 herp - Herbert Rosmanith
 *
 * forwarder functions
 *
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<string.h>
#include	<strings.h>
#include	<netinet/in.h>
#include	<netdb.h>
#include	<unistd.h>

#include	"glogin.h"
#include	"fwd.h"

/* client side functions */

int ntargets;		/* how many forwarding targets have been created */
int *chn2fwd=NULL;      /* fwd = chn2fwd[channel.id] */
u_short *fwd2chn=NULL;  /* channel.id = fwd2chn[fwd.fd] */
int nfds;

/* keeps indirect fd/channel mapping */

int fwd_init_table() {
        nfds=8*sizeof(fd_set);
        /* with 1024fds and int=4, this table is 4k */
        if ((chn2fwd=(int *)malloc(nfds))==NULL)
                return -1;
	/* and 8k for u_short entries */
        if ((fwd2chn=(u_short *)malloc(sizeof(u_short)*nfds))==NULL)
                return -1;
        memset(chn2fwd,0xff,nfds);
        memset(fwd2chn,0xff,sizeof(u_short)*nfds);
        return 0;
}

void fwd_create_listeners(int **ptrfwd,int nfd,int do_remote,int do_x11fwd,int x11fd,char **fwd_list) {
struct sockaddr_in s_in;
u_short port;
int *fwd;
const char *flag;
int i,j,k;

#ifdef DEBUG
        debug("> fwd_create_listeners alloc=%d\n",nfd+do_x11fwd+1);
	debug("  list follows:\n");
	for(i=0;i<nfd;i++)
		debug("  %s\n",fwd_list[i]);
	debug("  end of list\n");
#endif
	nfd+=do_x11fwd;
	if ((fwd=(int *)malloc((nfd+1)*sizeof(int)))==NULL)
		die("out of memory");
	debug("  created at %p\n",fwd);
	if (do_x11fwd)
		fwd[0]=x11fd;
        flag=do_remote?"-R":"-L";
        i=do_x11fwd;
        j=0;
        k=1;
        while(i<nfd) {
		debug("  fwd_create: i=%d nfd=%d\n",i,nfd);
                if ((fwd[i]=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
                        die("socket");
		debug("  fd=%d\n",fwd[i]);
                if (setsockopt(fwd[i],SOL_SOCKET,SO_REUSEADDR,&k,sizeof(k))==-1)
                        die("setsockopt");
		port=atoi(fwd_list[j++]);
                s_in.sin_family=AF_INET;
                s_in.sin_addr.s_addr=INADDR_ANY;
                s_in.sin_port=htons(port);
                if (bind(fwd[i],(struct sockaddr *)&s_in,sizeof(s_in))==-1)
                       	die("bind");
                if (listen(fwd[i],4)==-1)
                       	die("listen");
                i++;
        }
        debug("  i=%d\n",i);
        fwd[i]=-1;
	*ptrfwd=fwd;
}

void fwd_create_targets(fwd_desc_entry_t **ptrfwde,int nfd,int do_remote,int do_x11fwd,char **fwd_list) {
fwd_desc_entry_t *fwde;
const char *flag;
char *cp;
int i,j;

        debug("> fwd_create_targets nfd=%d fwd_list=%p\n",nfd,fwd_list);

	if (nfd && fwd_list==NULL)
		abort();

	if (nfd)
		debug("%s\n",fwd_list[0]);
	if (!do_remote)	// remote never has X11 in target table
		nfd+=do_x11fwd;
	if (nfd==0)
		return;
	if ((fwde=(fwd_desc_entry_t *)malloc(nfd*sizeof(fwd_desc_entry_t)))==NULL)
		die("out of memory");
	debug("  alloc= nfd=%d\n",nfd);
        flag=do_remote?"-L":"-R";
        i=do_remote?0:do_x11fwd;	/* no local X11 listener socket required */
	j=0;
	debug("  i=%d nfd=%d\n",i,nfd);
        while(i<nfd) {
	debug("  while i<nfd: i=%d nfd=%d\n",i,nfd);
                debug("  fwd_list[%d]=%s\n",j,fwd_list[j]);
                if ((cp=index(fwd_list[j],':'))==NULL)
                	die("illegal argument");
                fwde[i].host=++cp;
                if ((cp=index(cp,':'))==NULL)
                        die("illegal argument");
		fwde[i].service=strdup(cp+1);
                *cp=0;
                fwde[i].host=strdup(fwde[i].host);
                *cp=':';
                i++;
                j++;
        }
	*ptrfwde=fwde;
	ntargets=nfd;
}

#if 1
int connect_fwd(fwd_desc_entry_t *fwd) {
struct in_addr addr;
struct servent *se;
u_short port;

        debug("> connect_fwd(host=%s service=%s)\n",fwd->host,fwd->service);
        if (resolv(fwd->host,&addr)==-1)
                return -1;
        if ((port=atoi(fwd->service))==0) {
                if ((se=getservbyname(fwd->service,"tcp"))==NULL)
                	return -1;
                port=htons(se->s_port);
	}
	return connect_nb(addr,port);
}

#else 

int connect_fwd(fwd_desc_entry_t *fwd) {
struct sockaddr_in s_in;
struct servent *se;
u_short port;
int s=-1;
int i;

        debug("> connect_fwd(host=%s service=%s)\n",fwd->host,fwd->service);
        if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
                goto error_return;
	if (fcntl(s,F_SETFL,O_RDWR|O_NONBLOCK)==-1)
		goto error_return;
        if (resolv(fwd->host,&s_in.sin_addr)==-1)
                goto error_return;
        if ((port=atoi(fwd->service))==0) {
                if ((se=getservbyname(fwd->service,"tcp"))==NULL)
                	goto error_return;
                port=htons(se->s_port);
        }
        s_in.sin_port=htons(port);
        s_in.sin_family=AF_INET;
        if ((i=connect(s,(struct sockaddr *)&s_in,sizeof(s_in)))==-1
		&& errno!=EINPROGRESS)
                goto error_return;
        return s;
error_return:
	if (s!=-1)
		close(s);
	return -1;
}
#endif
