/*
 * Tue Oct 21 15:50:46 CEST 2003 herp - Herbert Rosmanith
 *                               hr@gup.uni-linz.ac.at
 *
 * glogin - grid login: a terminal session (tty) over GSS
 *
 * added features:
 *  o  VPNs with PPPD & glogin
 *  o  X11 forwarding
 *  o  local and remote TCP port-forwarding
 */

#include	"glogin.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/termios.h>
#include	<sys/signal.h>
#include	<sys/socket.h>
#include	<sys/stat.h>
#ifdef	HAVE_STROPTS
#include	<sys/stropts.h>
#endif
#include	<sys/ioctl.h>
#include	<sys/utsname.h>
#include	<sys/wait.h>
#include	<signal.h>
#include	<string.h>
#include	<strings.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<dirent.h>
#include	<fcntl.h>
#include	<pwd.h>
#include	<syslog.h>
#include	<netdb.h>
#include	<unistd.h>

#include	<globus_gass_server_ez.h>
#include	<globus_gram_client.h>
#include	<globus_gram_protocol.h>

#include	"channel.h"
#include	"fwd.h"
#include	"x11.h"

#define	GSH_MAX_DATA_SIZE	(16*1024-2)
//#define	GSH_MAX_DATA_SIZE	(16*1024)

char *self;
pid_t self_pid;
struct termios g_tc;
struct in_addr r_addr;	/* in addr of remote */
int eflag;		/* endian flag */
struct utsname uts;
struct hostent *self_hp;
struct in_addr self_addr;
char ghostname[MAXHOSTNAMELEN];
initcon_t g_ic;
char *cert_subject=NULL;

int enc_mode=1;  /* default=encrypt */

extern int ntargets;	/* from fwd.c: how many remote targets have been created */

#ifdef PROXY_DELEGATION
/* this proxy will allow hopping from host to host */
char x509_user_proxy[]="/tmp/x509up_XXXXXX";
time_t ctime_user_proxy=0;
#endif

static int do_winsize=0;
static pid_t cmdpid=0;
static char *fname_x509_up=NULL;

#ifdef DEBUG
void ls_dir(char *dir) {
struct dirent *de;
DIR *d;

	debug("=====\n|| %s ||\n=====\n",dir);
	if ((d=opendir(dir))==NULL) return;
	while((de=readdir(d))) {
		debug("%s\n",de->d_name);
	}
	closedir(d);
}
#endif

char *opt_rsl=NULL;
char *opt_queue=NULL;
char *opt_outfile=NULL;
char *opt_infile=NULL;
char *opt_errsock=NULL;
u_int opt_port=0;
int do_addr=0;
char *opt_addr=NULL;
int do_fork=0;
int do_remote=0;
int do_restore=0;
int do_encrypt=1;
int do_nocmd=0;
static int do_pty=0;
int do_x11fwd=-1;
int do_staging=0;
int do_exec=0;

void send_winsize(int s,int ttyfd) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+sizeof(struct winsize);
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_WINSIZE;
	/* if we don't have a WINSIZE, then this is
	 * possibly a programm pipe: just ignore the
	 * error message. send an zero-winsize instead.
	 */
	if (ioctl(ttyfd,TIOCGWINSZ,&tmsg->data)==-1)
		memset(tmsg->data,0,sizeof(struct winsize));
	gsswrite(s,&ibuf);
}

// obsolete
#if 0
void send_term_cc(int s) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+NCC;
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_TERM_CC;
	memcpy(&tmsg->data,&g_tc.c_cc,NCC);
	gsswrite(s,&ibuf);
}
#endif

void send_cookie(int s) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+sizeof(xcookie);
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_XCOOKIE;
	memcpy(&tmsg->data,xcookie,sizeof(xcookie));
	gsswrite(s,&ibuf);
}

void send_dsploc(int s) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;
int i;

	debug("> send_dsploc xdisploc=%s\n",xdisploc);

        i=strlen(xdisploc)+1;
        ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+strlen(xdisploc)+1;
        chn=(struct chan_s *)alloca(ibuf.length);
        chn->id=CHNID_TMSG;
        ibuf.value=chn;
        tmsg=(struct tmsg_s *)chn->data;
        tmsg->type=TMSG_XDSPLOC;
        memcpy(&tmsg->data,xdisploc,i);
        gsswrite(s,&ibuf); 
}

void send_fwd_connect(int s,int findex,int fd) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;
struct tmsg_fwd_connect *tfc;

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+sizeof(findex)+sizeof(fd);
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_FWD_CONNECT;
	tfc=(struct tmsg_fwd_connect *)&tmsg->data;
	tfc->findex=findex;
	tfc->fd=fd;
	gsswrite(s,&ibuf);
}

void send_fwd_reply(int s,int lfd,int rfd) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;
struct tmsg_fwd_reply *tcr;

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+sizeof(lfd)+sizeof(rfd);
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_FWD_REPLY;
	tcr=(struct tmsg_fwd_reply *)&tmsg->data;
	tcr->lfd=lfd;
	tcr->rfd=rfd;
	gsswrite(s,&ibuf);

}

#ifdef	PROXY_DELEGATION
void send_proxy(int s) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;
//struct tmsg_proxy *tpx;
char pname[]="/tmp/x509up_XXXXXX";
struct stat st;
int fd;

	sprintf(pname+12,"u%d",getuid());
	if (stat(pname,&st)==-1)
		die("stat");
	//ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+sizeof(int)+st.st_size;
	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type)+st.st_size;
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_PROXY;
	//tpx=(struct tmsg_proxy *)&tmsg->data;
	//tpx->len=st.st_size;
	if ((fd=open(pname,O_RDONLY))==-1)
		die("open");
	//read(fd,tpx->data,st.st_size);
	read(fd,tmsg->data,st.st_size);
	close(fd);
	gsswrite(s,&ibuf);
}
#endif

void send_eot(int s) {
gss_buffer_desc ibuf;
struct chan_s *chn;
struct tmsg_s *tmsg;
static int sent=0;

	if (sent)
		return;
	sent=1;

	debug("send_eot s=%d\n",s);

	ibuf.length=CHN_HDRLEN+sizeof(tmsg->type);
	chn=(struct chan_s *)alloca(ibuf.length);
	chn->id=CHNID_TMSG;
	ibuf.value=chn;
	tmsg=(struct tmsg_s *)chn->data;
	tmsg->type=TMSG_EOT;
	gsswrite(s,&ibuf);
}

void send_error_msg(int s,char *msg) {
gss_buffer_desc ibuf;
struct chan_s *chn;

	debug("send_error_msg msg=%s\n",msg);

        ibuf.length=CHN_HDRLEN+strlen(msg)+1;
        chn=(struct chan_s *)alloca(ibuf.length);
        chn->id=CHNID_STDERR;
        strcpy(chn->data,msg);
        ibuf.value=chn;
        gsswrite(s,&ibuf);
}

#if 0
void sync_eot(int s,int flag) {
char buf[1024];
u_short sh;
int i;

        debug("syncing eot");
        if (!flag)
                send_eot(s);
        sh=0xffff;
        write(s,&sh,2);
        for(;;) {
                if (read(s,&sh,2)!=2)
                        return;
                if (sh==0xffff)	/* we are synced */
                        return;
                /* eat up all other messages */
                if (eflag) sh=swap16(sh);
		while(sh) {
                        if ((i=read(s,&buf,sh))<1)
                                return;
			sh-=i;
		}
        }
        return;
}
#endif

#if 0
void sigchld(int nsig) {
	debug("got SIGCHLD\n");
}
#endif

static inline void ua_int16(void *p1,void *p2) {
#if defined(__sun__)
char *dst=(char *)p1;
char *src=(char *)p2;
        memcpy(dst,src,2);
#else
*(short *)p1=*(short *)p2;
#endif
}

static inline void ua_int32(void *p1,void *p2) {
#if defined(__sun__)
char *dst=(char *)p1;
char *src=(char *)p2;
        memcpy(dst,src,4);
#else
*(long *)p1=*(long *)p2;
#endif
}

gss_buffer_desc ibuf;
struct chan_s *ichn,*ochn;

int nrfwc=0;	/* number of established forwarded connections */
fd_set rfds;	/* the read filedescriptor set */
fd_set wfds;	/* the write filedescript set (for non blocking connects) */
fd_set *pwfds=NULL;

int handle_tunnel_socket(int *n,int sockfd,int ttyfd,int tty_outfd,int tty_errfd,
	fwd_desc_entry_t *fwde,int *done) {
gss_buffer_desc wbuf;
struct chan_s *ochn;
u_short chnid;
OM_uint32 min;
int more;
int xa;
int i;

	debug("/* incoming message from socket */ (done=%d)\n",done);
again:
	debug("   again:\n");
	i=gssread(sockfd,&wbuf,&more);
	debug("   again: more=%d\n",more);
	if (i==-1) {
		debug("EOF at tunnel socket\n");
		*done=1;
		return 0;	/* 0->break; */
	}
	if (i==0)
		return 0;
	if (wbuf.length==0) {
		debug("dazed and confused, but trying to continue: null message\n");
		return 1;	/* 1->continue; */
	}
	ochn=(struct chan_s *)wbuf.value;
	ua_int16(&chnid,&ochn->id);	// sigbus on sparc et. al.
	if (eflag)
		chnid=swap16(chnid);
#ifdef DEBUG
	debug("got %d bytes of data on channel %d\n",
		wbuf.length,chnid);
	if (wbuf.length-CHN_HDRLEN==0)
		debug("ZERO-sized packet on channel %d\n",chnid);
#endif
	{
	u_char *cp=(u_char *)ochn;
		for(i=0;i<wbuf.length;i++) {
			debug("  %.2x %c\n",*cp,*cp);
			cp++;
		}
	}
	switch(chnid) {
	/* a message for the shell */
	case CHNID_STDINOUT:
	case CHNID_STDERR:
		debug("msg for channel %d\n",chnid);
		if (wbuf.length-CHN_HDRLEN) {
			write(chnid==CHNID_STDINOUT?tty_outfd:tty_errfd,
				ochn->data,wbuf.length-CHN_HDRLEN);
#ifdef DEBUG
			ochn->data[wbuf.length-CHN_HDRLEN]=0;
			debug("data=%s",ochn->data);
#endif
		}
		else {
			/* packet size zero means EOF */
			debug("terminal shell is dead\n");
			/* remove our reading fds from the fd-set, but
			 * keep them open, so we still can write to them,
			 * but dont accept new input.
			 */
			FD_CLR(ttyfd,&rfds);
			if (tty_errfd!=-1)
				FD_CLR(tty_errfd,&rfds);
			/* if we have active forwarded connections, we
			 * must not terminate. wait until all forwarded
			 * connections have terminated
			 */
			debug("nrfwc=%d\n",nrfwc);
			if (nrfwc==0) /*done=1;*/
				;
			else if (!do_remote) printf("%d forwarded connection(s) still open...\r\n",nrfwc);
		}
		break;
	/* a message for terminal control */
	case CHNID_TMSG: {
	struct tmsg_s *tmsg=(struct tmsg_s *)ochn->data;
	u_short mtyp;
		ua_int16(&mtyp,&tmsg->type);	// sigbugs sparc/sgi ...
		if (eflag) mtyp=swap16(mtyp);
		switch(mtyp) {
		case TMSG_WINSIZE:
			if (eflag)
				swapws((struct winsize *)&tmsg->data);
			ioctl(ttyfd,TIOCSWINSZ,&tmsg->data);
			break;
// obsolete
#if 0
		case TMSG_TERM_CC:
			/* the TCGETS is required, since some
			 * program might have changed the termios
			 * of course that's just a HACK which needs
			 * to be FIXED: TERM_CC should only be allowed
			 * at the start of the program.
			 */
			ioctl(ttyfd,TCGETS,&g_tc);
			memcpy(&g_tc.c_cc,&tmsg->data,NCC);
			ioctl(ttyfd,TCSETS,&g_tc);
			break;
#endif
// same situation here (to be done!)
		case TMSG_XCOOKIE: {
		char xcookie_str[33];
			bin2hex(xcookie_str,tmsg->data,16);
			debug("got xcookie %s\n",xcookie_str);
			x11_add_Xauthority(xdisploc,xcookie_str);	/* pf_inet */
			x11_add_Xauthority(xunixloc,xcookie_str);	/* pf_unix */
			/* only now it is save to send the display
			 * location. this avoids a potential race
			 * condition
			 */
			send_dsploc(sockfd);
			break;
		}
		case TMSG_XDSPLOC: {
		char xcookie_str[33];
		char *cp;
			debug("got xdsploc %s\n",tmsg->data);
			if ((cp=index(tmsg->data,':'))==NULL)	/* bah */
				break;
			bin2hex(xcookie_str,xcookie,16);
			if ((xdisploc=strdup(tmsg->data))==NULL)
				die("out of memory");
			x11_add_Xauthority(xdisploc,xcookie_str);
			break;
		}
		case TMSG_FWD_CONNECT: {
		struct tmsg_fwd_connect *tfc;
		int ix,fd;
			// XXX - alignment issues
			tfc=(struct tmsg_fwd_connect *)tmsg->data;
			ua_int32(&ix,&tfc->findex);
			ua_int32(&fd,&tfc->fd);
			if (eflag) {
				ix=swap32(ix);
				fd=swap32(fd);
			}
			debug("got fwd_connect rfindex=%d rfd=%d\n",ix,fd);
			if (ix>=ntargets)
				die("forwarder out of range");
			if (!do_remote && do_x11fwd && ix==0)	/* this is the X11-forwarder */
				xa=x11_open_display();
			else xa=connect_fwd(&fwde[ix]);
			debug("  xa=%d (established forwarder fd)\n",xa);
			/* failed connection to forwarder:
			 * send a negative response or
			 * close the channel
			 */
			if (xa==-1) {	/* failed */
				if (do_remote)
					send_fwd_reply(sockfd,fd,-1);
				else {
					// XXX - TMSG_EOF bzw. TMSG_EOT waere besser
					// die("XXX");
					ichn->id=tfc->fd;   /* send eof */
					ibuf.length=CHN_HDRLEN;
					ibuf.value=ichn;
					gsswrite(sockfd,&ibuf);
				}
			}
			else {
				debug("new target fd xa=%d tfc->fd=%d\n",
					xa,fd);
				chn2fwd[fd]=xa;
				fwd2chn[xa]=fd;
				FD_SET(xa,&rfds);
				FD_SET(xa,&wfds);
				pwfds=&wfds;
				if (xa>*n)
					*n=xa;
				if (do_remote)
					send_fwd_reply(sockfd,fd,xa);	/* we have to swap fds here */
				nrfwc++;
				debug("incr: nrfwc=%d\n",nrfwc);
			}
			break;
		}
		case TMSG_FWD_REPLY: {
		struct tmsg_fwd_reply *tcr;
		int lfd,rfd;
			tcr=(struct tmsg_fwd_reply *)tmsg->data;
			ua_int32(&lfd,&tcr->lfd);	// unaligned access 
			ua_int32(&rfd,&tcr->rfd);	// unaligned access
			if (eflag) {
				lfd=swap32(lfd);
				rfd=swap32(rfd);
			}
			debug("got fwd_reply lfd=%d rfd=%d\n",lfd,rfd);
			if (rfd==-1) { /* something failed */
				close(lfd);
				nrfwc--;
				debug("decr: nrfwc=%d\n",nrfwc);
				if (nrfwc==0 && !do_nocmd && !FD_ISSET(ttyfd,&rfds))
					*done=1;
			}
			else {
				chn2fwd[rfd]=lfd;
				fwd2chn[lfd]=rfd;
				FD_SET(lfd,&rfds);
				if (lfd>*n)
					*n=lfd;
			}
			break;
		}
#ifdef	PROXY_DELEGATION
		case TMSG_PROXY: {
		// struct tmsg_proxy *tpx;
		struct stat st;
		int fd;
		int i;
			sprintf(x509_user_proxy+12,"u%d",getuid());
			/* a proxy already exists, leave it alone */
			if (stat(x509_user_proxy,&st)==0)
				break;
			// tpx=(struct tmsg_proxy *)tmsg->data;
			// if (eflag)
			// 	tpx->len=swap32(tpx->len);
			// if ((fd=open(pname,O_CREAT|O_WRONLY|O_TRUNC|O_EXCL,0600))==-1)
			if ((fd=open(x509_user_proxy,O_CREAT|O_WRONLY|O_TRUNC,0600))==-1)
				die("open");
			//write(fd,tpx->data,tpx->len);
			i=wbuf.length-CHN_HDRLEN-sizeof(tmsg->type);
			//write(fd,tpx->data,i);
			write(fd,tmsg->data,i);
			close(fd);
			/* remember ctime */
			if (stat(x509_user_proxy,&st)==-1)
				die("stat");
			ctime_user_proxy=st.st_ctime;
			break;
		}
#endif
#if 1
		case TMSG_EOT:
			//if (!do_remote)
			//	fprintf(stderr,"got EOT\n");
			debug("got EOT\n");
			*done=1;
			if (do_remote)
				die("TMSG_EOT consistency failure");
			debug("local sending eot\n");
			send_eot(sockfd);

			break;
#endif
		default:
			debug("unprocessed TMSG\n");
			// XXX if we want to log and continue
			// processing, we'd have to reserve an
			// fd
			// logger(LOG_ERR,"unknown TMSG\n");
			// exit(-1);
			break;
		}
		break;
	}
	/* a message for a forwarded connection */
	default:
		debug("/* a message for a forwarded connection */");
		/* remote: direct mapping, local: table lookup */
		xa=do_remote?chnid:chn2fwd[chnid];
		debug("  forwarded fd=%d\n",xa);
		if (xa>-1) {
			if (wbuf.length>CHN_HDRLEN) {
				debug("  writing %d bytes of data to fd %d\n",
					wbuf.length-CHN_HDRLEN,xa);
				write(xa,ochn->data,wbuf.length-CHN_HDRLEN);
			}
			else {	/* zero packet: close connection */
				debug("removing connection fd %d\n",xa);
				FD_CLR(xa,&rfds);
				close(xa);
				if (xa==*n)
					*n=maxfd(*n,&rfds);
				/* remove from tables */
				if (!do_remote) {
					chn2fwd[chnid]=-1;
					fwd2chn[xa]=(u_short)-1;
				}
				nrfwc--;
				debug("decr: nrfwc=%d\n",nrfwc);
				if (nrfwc==0 && !do_nocmd && !FD_ISSET(ttyfd,&rfds))
					*done=1;
				// debug("removed forwarded connection, fd%d\n",xfd);
				/* don't inform if we still have a shell */
				// if (!FD_ISSET(ttyfd,&rfds)) {
				// 	if (nrfwc==0) *done=1;
				// 	else if (!do_remote) printf("%d forwarded connection(s) still open...\r\n",nrfwc);
				// }
			}
		}
		else die("illegal channel");
		break;
	}
	if (do_encrypt)
		gss_release_buffer(&min,&wbuf);
	if (more)
		goto again;
	return 1;
}

void handle_tty_fd(int fd,int sockfd,int ttyfd,int *done) {
static int nreof=0;
int i;

	//i=read(ttyfd,ichn->u.data,ichn->len);

	// XXX - channel-id und daten im clear-text mode
	// um 5 byte nach hinten, dann kann das laengen feld
	// vorne rein und das writev() koennte durch ein write()
	// ausgetauscht werden. stellt sich allerdings die frage,
	// ob das wirklich was bringt.

	//i=read(fd,&ichn->data,4096);
	//i=read(fd,&ichn->data,8192);
	//i=read(fd,&ichn->data,16384);
	//i=read(fd,&ichn->data,32768);
	//i=read(fd,&ichn->data,65536);

	i=read(fd,&ichn->data,GSH_MAX_DATA_SIZE);

#ifdef DEBUG
	ichn->data[i>=0?i:0]=0;
	debug("/* message from the shell (%s) */ %d bytes %s\n",fd==ttyfd?"stdout":"stderr",i,ichn->data);
	if (i>0) {
		ichn->data[i]=0;
		//debug("data=%s\n",ichn->data);
		write(254,ichn->data,i);
	}
#endif
	/* determine whether we are really done.
 	* the point here is that even though one
 	* std-fd may return 0 or -1, the other one
 	* might still have data. the *real* eof
 	* condition is indicated by waitpid returning
 	* -1
	*
	* XXX - this is not true! the *real* eof
	*       condition can only be found by
	*       counting how often a zero occured
	*       on reading - two times indicates
	*	eof.
 	*/
	//if (i<=0) {
	if (i==0) {
		/* if we have a fifo, allow restarting
		 * IO on it.
		 */
                if (opt_infile) {
                struct stat st;
                        if (fstat(fd,&st)==-1)
                                die("fstat");
                        if (S_ISFIFO(st.st_mode)) {     // it's a fifo, reopen it
                                if ((i=open(opt_infile,O_RDONLY|O_NONBLOCK))==-1)
                                        die("open");
                                dup2(i,fd);
                                close(i);
                                return;
                        }
                }
		FD_CLR(fd,&rfds);	/* remove it */
		if (i==-1) {		/* an error occured */
			debug("i=-1 %d %s\n",errno,strerror(errno));
		}
		else {
			nreof++;
			if (nreof<2)
				return;
		}
		if (nrfwc==0)
			*done=1;
	}
	else if (i>0) {
		ichn->id=fd==ttyfd?CHNID_STDINOUT:CHNID_STDERR;
		//ibuf.length=(i>0?i:0)+CHN_HDRLEN;
		ibuf.length=i+CHN_HDRLEN;
		ibuf.value=ichn;
		debug("  gsswrite %d bytes of data to channel %d\n",
			ibuf.length,ichn->id);
		gsswrite(sockfd,&ibuf);
	}
	else if (i==-1) {
		debug("shell is terminating\n");
		FD_CLR(fd,&rfds);
		if (nrfwc==0)
			*done=1;

		/* closing fd is problematic, since it will
	 	* dig a "hole" in  the lower fd-map! this
	 	* will lead to serious problems with the
	 	* forwarders. do not close it!
		*
		* XXX - a better solution is to create
		*       a fdset which contains all forwarders
	 	*/

		//close(fd);

		// if we dont close the fd, then setting tty/err/fd to -1
		// is not required, in contrast, will lead to errors only.

		//if (fd==ttyfd)
		//	ttyfd=tty_errfd=-1; /* invalidate, since subsequent open() could reuse this fd */
	}
}

void handle_forwarder_creation(int *n,int fd,int sockfd,int *fwdfd) {
struct sockaddr_in saddr;
socklen_t len;
int i,xa;

	/* incoming message from forwarder */
	debug("/* incoming message from forwarder */\n");
	len=sizeof(saddr);
	if ((xa=accept(fd,(struct sockaddr *)&saddr,&len))==-1) {
		//perror("accept");	// XXX WARN
		debug("accept on fd=%d failed\n",fd);
		die("accept");	// XXX WARN
		//continue;
	}
        i=128*1024;
        if (setsockopt(xa,SOL_SOCKET,SO_SNDBUF,&i,sizeof(i))==-1)
                die("setsockopt");
        if (setsockopt(xa,SOL_SOCKET,SO_RCVBUF,&i,sizeof(i))==-1)
                die("setsockopt");

	/* XXX if we have a large forwarder table
	 * we should better use a hash-table. but it's
	 * unlikely that we have a large forwarder table
	 * linear search is fast enough. new connections
	 * usually are not created at a high rate.
	 */
	for(i=0;fwdfd[i]>-1;i++)	/* get index */
		if (fwdfd[i]==fd)
			break;		// i now contains the forwarder index

	/* on the local system, we have to ask the remote to create a tuple for us */
	/* on the remote system, we can instantly create the forwarding tuple */
	/* on the local system, we have to search the index first */
	if (do_remote) {
		debug("new forwarded connection: fd=%d\n",xa);
		FD_SET(xa,&rfds);
		if (xa>*n)
			*n=xa;
	}

	/* tell the other end that xa has been created at index i */
	send_fwd_connect(sockfd,i,xa);
	nrfwc++;
	debug("incr: nrfwc=%d\n",nrfwc);
}

//void passthru(int ttyfd,int tty_outfd,int sockfd,int x11fd,int fwdfd) {
void passthru(int ttyfd,int tty_outfd,int tty_errfd,int sockfd,int *fwdfd,fwd_desc_entry_t *fwde) {
//gss_buffer_desc ibuf,wbuf;
//struct sockaddr_in saddr;
//socklen_t len;
//fd_set rfds,tfds;
fd_set tfds;
fd_set listen_fds;
//int xa,fd;
int fd;
int maxfwd;
int i,j,n;
int done=0;

	debug("> passthru %s ttyfd=%d tty_outfd=%d tty_errfd=%d sockfd=%d\n",
			do_remote?"REMOTE":"LOCAL",
			ttyfd,tty_outfd,tty_errfd,sockfd);

	FD_ZERO(&listen_fds);
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_SET(sockfd,&rfds);	/* sockfd = remote filedescriptor */
	
	if (ttyfd>-1)
		FD_SET(ttyfd,&rfds);	/* ttyfd = local filedescriptor */

	/* remote glogin _reads_ from tty_errfd, while
 	* local glogin _writes_ to tty_errfd.
 	*/
	if (do_remote && tty_errfd>-1)
		FD_SET(tty_errfd,&rfds);

	/* calculate the maximum fd from which we
 	* read data. also enable all fds in the rfds,
 	* so that select() notices IO on them.
 	*/
	n=ttyfd>sockfd?ttyfd:sockfd;
	n=tty_errfd>n?tty_errfd:n;
	for(i=0;;i++) {
		if ((j=fwdfd[i])==-1)
			break;
		debug("fwdfd[%d]=%d\n",i,j);
		FD_SET(j,&listen_fds);
		FD_SET(j,&rfds);
		n=j>n?j:n;
	}
	/* if there is no forwarder, i will be zero */
	maxfwd=i?fwdfd[i-1]:-1;

	fwd_init_table();

	//ichn=(struct chan_s *)alloca(5+CHN_HDRLEN+4096);
	//ichn=(struct chan_s *)alloca(CHN_HDRLEN+4096);
	//ichn=(struct chan_s *)alloca(CHN_HDRLEN+8192);
	//ichn=(struct chan_s *)alloca(CHN_HDRLEN+16384);
	//ichn=(struct chan_s *)alloca(CHN_HDRLEN+32768);
	//ichn=(struct chan_s *)alloca(CHN_HDRLEN+65536);
	ichn=(struct chan_s *)alloca(CHN_HDRLEN+GSH_MAX_DATA_SIZE);
	ibuf.value=ichn;
	if (!do_remote) {	/* local term tells remote our settings */
		// already sent initially, and then it never
		// changes
		//send_term_cc(sockfd);
		// not neccessary too, however, it might change.
		//send_winsize(sockfd,ttyfd);
		if (do_x11fwd) {
			debug("sending x11 cookie\n");
			send_cookie(sockfd);
		}
#ifdef	PROXY_DELEGATION
		send_proxy(sockfd);
#endif
	}
	//else ioctl(ttyfd,TCGETS,&g_tc); /* have that ready for later use */
	// else if (more)
	// 	handle_tunnel_socket(&n,sockfd,ttyfd,tty_outfd,tty_errfd,fwde,&done);

extern int token_buffer_empty;

debug("is token_buffer empty?: %d\n",token_buffer_empty);

	// ugly. on the other hand, it's only needed for the first time

	if (!token_buffer_empty) {
debug("TOKENBUffer NOT empty\n");
		FD_ZERO(&tfds);
		FD_SET(sockfd,&tfds);
		goto past_select;
	}

	while(!done) {
		tfds=rfds;
		debug("> select n=%d\n",n);
		//i=select(n+1,&tfds,&wfds,NULL,NULL);
		i=select(n+1,&tfds,pwfds,NULL,NULL);
		if (i==-1) {
			if (errno!=EINTR)
				die("select");
			debug("select got EINTR\n");
			if (do_winsize) {
				send_winsize(sockfd,ttyfd);
				do_winsize=0;
			}
			continue;
		}
past_select:
#if 1
		// pointer to write fdset will
		// contain all fds waiting for NB connect()
		if (pwfds) {
		int all_clear;
		int nb=(n>>3)+1;

			debug("enter for wfds: n=%d\n",n);
			for(fd=0;fd<=n;fd++) {
				if (!FD_ISSET(fd,&wfds))
					all_clear=0;
				else {
					debug("connect fd=%d alive\n",fd);
					/* go blocking by clearing O_NONBLOCK flag */
					if (fcntl(fd,F_SETFL,O_RDWR)==-1)
						die("fcntl");
					FD_CLR(fd,&wfds);
				}
			}
			/* if there are no fds, we dont need pwfds any more */
			all_clear=1;
			for(i=0;i<nb;i++)
				if (((char *)pwfds)[i])
					all_clear=0;
			if (all_clear)
				pwfds=NULL;
		}
#endif
		debug("enter for rfds: n=%d\n",n);
		for(fd=0;fd<=n;fd++) {
		int eot=0;
			if (!FD_ISSET(fd,&tfds))
				continue;
			debug("message from fd%d\n",fd);
			debug("  sockfd=%d\n",sockfd);
			/* incoming message from terminal */
			if (fd==ttyfd || fd==tty_errfd) {
				handle_tty_fd(fd,sockfd,ttyfd,&eot);
				if (do_remote && eot)
					done=1;	// leave loop
#if 0
				if (do_remote && eot) {
					debug("remote sending eot\n");
					send_eot(sockfd);
				}
#endif
				continue;
			}

			/* incoming message from socket */
			if (fd==sockfd) {
				if (handle_tunnel_socket(&n,sockfd,ttyfd,tty_outfd,tty_errfd,fwde,&done)==0)
					break;
				continue;
			}

			/*
			 * if fd is <= than maxfwd, then we have a
			 * listener(forwarder), because the specific
			 * tty-fds have already been handled above.
			 * if fd is  > than maxfwd, then we have a
			 * forwarded connection. (next case below)
			 *
			 * note: I don't know if it's better to use
			 * an fdset for this. I could set all created
			 * forwarded connections to one and ask
			 *  if (!FD_ISSET(fd,&create_fds)) { ...
                         *
                         * Fri Jan 19 12:50:50 CET 2007 herp
                         * - ok, I've implemented it this way now.
                         *
			 */
			//if (fd<=maxfwd) {
			if (FD_ISSET(fd,&listen_fds)) {
				handle_forwarder_creation(&n,fd,sockfd,fwdfd);
				continue;
			}

			/* default - incoming message from an established forwarded connection */
			debug("/* default - incoming message from a client */\n");

			//i=read(fd,&ichn->data,4096);
			//i=read(fd,&ichn->data,8192);
			//i=read(fd,&ichn->data,16384);
			//i=read(fd,&ichn->data,32768);
			//i=read(fd,&ichn->data,65536);

			i=read(fd,&ichn->data,GSH_MAX_DATA_SIZE);
			if (i==-1)
				i=0;
			/* local glogin:  lookup in fd->channel map
			 * remot glogin: direct mapping fd->channel
			 */
			ichn->id=do_remote?fd:fwd2chn[fd];
			ibuf.length=i+CHN_HDRLEN;
			ibuf.value=ichn;
			debug("sending %smessage for fd=%d on channel %d\n",
				i==0?"CLOSE ":"",fd,ichn->id);
			gsswrite(sockfd,&ibuf);
			if (i==0) { /* EOF on client */
				FD_CLR(fd,&rfds);
				if (fd==n) n=maxfd(n,&rfds);
				close(fd);
				if (!do_remote) {	/* remove from tables */
					fwd2chn[fd]=-1;
					chn2fwd[ichn->id]=-1;
				}
				debug("removed fd%d, new maxfd=%d\n",fd,n);
				nrfwc--;
				debug("decr: nrfwc=%d\n",nrfwc);
				if (nrfwc==0 && !do_nocmd && !FD_ISSET(ttyfd,&rfds)) {
					debug("-- we are done --\n");
					done=1;
				}
			}
		} /* for */
	} /* while(!done) */

#if 0
	/* synchronise end of transmission */
	sync_eot(sockfd,seen_eot);
#endif
	/*
	 * end-of-transmission works as follows:
	 * both partners have to agree on eot, so:
	 * 1st, remote sends an eot but continues processing.
	 * 2nd, local sends an eot when receiving eot,
	 *      and stops processing.
	 * 3rd, remote gets an eot and stops processing.
	 *
	 * note: only the remote knows when to terminate
	 * the transmission (local is only a dumb terminal),
	 * therefore the remote will initiate termination
	 * and only the remote will send the first eot.
	 */
	if (do_remote) {
	gss_buffer_desc wbuf;
	OM_uint32 min;
		debug("remote sending eot\n");
		send_eot(sockfd);
		FD_SET(0,&tfds);
		select(1,&tfds,NULL,NULL,NULL);
		gssread(0,&wbuf,NULL);
		if (do_encrypt)
			gss_release_buffer(&min,&wbuf);
	}
	debug("< passthru\n");
}

#if 0
void write_version(int s) {
char buf[32];

	debug("sending version=%s\n",GSH_PROTOCOL);

	if (strlen(GSH_PROTOCOL)>sizeof(buf)-4)	/* bugger */
		die("strange protocol");
	buf[0]=0xff;
	buf[1]=0xfa;
	buf[2]=strlen(GSH_PROTOCOL)+1;
	strcpy(buf+3,GSH_PROTOCOL);
	write(s,buf,buf[2]+3);
}
#endif

int check_proto_version(int s) {
u_char buf[32];
int i;

        debug("> check_proto_version\n");
        if ((i=read(s,buf,32))<4) {
                return -1;
        }
        if (buf[0]!=0xff && buf[1]!=0xfa) {
                fprintf(stderr,"couldn't read protocol signature\n");
                return -1;
        }
        if (buf[2]+3!=i) {
                fprintf(stderr,"couldn't read protocol version\n");
                return -1;
        }
        if (strcmp((char *)buf+3,GSH_PROTOCOL)) {
                fprintf(stderr,"wrong protocol version\n");
                return -1;
        }
        debug("  proto_version is %s\n",buf+3);
        return 0;
}

#ifndef I2G
/*
 * lcgpbs modfies $HOME - xauth doesnt like that.
 */
void chgenv_cd_home() {
struct passwd *pw;
char buf[256]={"HOME="};
char *cp;

	if ((pw=getpwuid(getuid()))==NULL)
		return;
	if ((cp=getenv("HOME")) && strcmp(cp,pw->pw_dir)) {
		strncpy(buf+5,pw->pw_dir,sizeof(buf)-5);
		putenv(strdup(buf));	// possible memory leak - but see manpage
		chdir(pw->pw_dir);
	}
	endpwent();
}
#endif

/*
 * create remote child process
 */

void execv_cmd(char *cmd,initcon_t *is,char *pts,int ptm,int pup[2],int pdn[2],int per[2]) {
char *cp;
int fd;
struct sockaddr_in s_in,p_in;
socklen_t len;
rcmd_t *rcmd;
#ifdef DEBUG
int i;
#endif

#ifndef I2G
	chgenv_cd_home();	// needed for lcgpbs (nfs mounted, $HOME modified)
#endif

	switch((cmdpid=fork())) {
	case -1:
		die("fork");
	case 0:
		/* for GSH_CONNECTION */
		len=sizeof(p_in);
		if (getpeername(0,(struct sockaddr *)&p_in,&len)==-1)
			die("getpeername");
		len=sizeof(s_in);
		if (getsockname(0,(struct sockaddr *)&s_in,&len)==-1)
			die("getsockname");
		// XXX
		// XXX - nachlesen: wofuer genau ist eine ``session'' gut?
		// XXX              was ist der unterschied zwischen ssn.
		// XXX              process group?
		// XXX   (stevens)
		// XXX
		if (do_pty) {
#if 1
			if (setsid()==-1)
				die("setsid");
			// unsetenv("HOME");
			// chdir("/");
#endif
		// if (do_pty) {
			/* this section definitely works */
			if ((fd=open(pts,O_RDWR))==-1)
				die("open pts");
			if (ioctl(fd,TCGETS,&g_tc)==-1)
				die("ioctl");
			memcpy(g_tc.c_cc,is->c_cc,NCC);
			if (ioctl(fd,TCSETS,&g_tc)==-1)
				die("ioctl");
			if (eflag)
				swapws(&is->ws);
			if (ioctl(fd,TIOCSWINSZ,&is->ws)==-1)
				die("ioctl");
#ifdef I_PUSH
                        if (ioctl(fd,I_PUSH,"ptem")==-1)
                                die("I_PUSH ptem");
                        if (ioctl(fd,I_PUSH,"ldterm")==-1)
                                die("I_PUSH ldterm");
                        if (ioctl(fd,I_PUSH,"ttcompat")==-1)
                                die("I_PUSH ttcompat");
#endif
			/* redirect stdX to pty */
			if (fd!=0) dup2(fd,0);
			if (fd!=1) dup2(fd,1);
			if (fd!=2) dup2(fd,2);
		}
		else {
#if 0
			         readfd write  readerr

	passthru(pup[0],pdn[0],per[0],as,fwdfd,fwde);
			up1=write
			dn1=read
			er1=writerr
#endif
			// order is important. first dup(pup), then
			// dup(pdn), then dup(per) since:
			// pup0=1 pup1=2 pdn0=3 pdn1=4 per0=5 per1=6
			// which results from order of pipe-creation.

			dup2(pup[1],1);
			dup2(pdn[0],0);
			dup2(per[1],2);
		}
		rcmd=rcmd_split(cmd,strcmp(cmd,"")==0);
		if (do_x11fwd)
			if ((cp=(char *)malloc(/*DISPLAY=localhost:xxx\0*/22))) {
				sprintf(cp,"DISPLAY=localhost:%d",xdispno);
				putenv(cp);
			}

		/* various envvars */
		putenv("TERM=vt100");	/* see comment in ishd.c */

		/* if we dont have a tty, GSH_TTY makes no sense and
		 * will crash due to pts==NULL anyway. therefore ...
		 */
		if (do_pty && (cp=(char *)malloc(/*GSH_TTY=*/8+strlen(pts)+1))) {
				strcpy(cp,"GSH_TTY=");
				strcat(cp,pts);
				putenv(cp);
			}
		/* GSH_CONN. is max. 59 bytes. give 'em all they would ever want */
		if ((cp=(char *)malloc(/*GSH_CONNECTION=inaddr port inaddr port*/59))) {
			sprintf(cp,"GSH_CONNECTION=%s %hu ",
				inet_ntoa(p_in.sin_addr),
				ntohs(p_in.sin_port));
			/* this is required since inet_ntoa uses a static buffer */
			sprintf(cp+strlen(cp),"%s %hu",
				inet_ntoa(s_in.sin_addr),
				ntohs(s_in.sin_port));
			putenv(cp);
		}

		if ((cp=(char *)malloc(24))) {
			snprintf(cp,24,"GSH_PPID=%d",getppid());
			putenv(cp);
		}
#ifdef DEBUG
		debug("execvp(rcmd->cmd=%s)\n",rcmd->cmd);
		debug("rcmd->nargs=%d\n",rcmd->nargs);
		for(i=0;rcmd->args[i];i++)
			debug("rcmd->args[%d]=%s\n",i,rcmd->args[i]);
#endif
		closeall_but(0,1,2,-1);
#ifndef I2G
		chdir(rcmd->pw->pw_dir);
#endif
#ifdef I2G
		chmod(rcmd->cmd,0550);
#endif
		execvp(rcmd->cmd,rcmd->args);

		close(1);	/* no idea what's going on there ... */
		die("execvp");	/* ... but we get crappy output from */
				/*     globus if we dont close fd1   */

	default:
		break;
	}
	//chdir("/");	// dont - xauth might fail
	debug("cmdpid=%d\n",cmdpid);

#if 1
	if (do_pty) {
		pup[0]=ptm;	// one for read
		pdn[1]=ptm;	// one for write
		per[0]=-1;	/* makes no sense for a pty */
	}
	else {
		// note: it doesnt matter if there are "holes" in
		// the range of filedescriptors now, since
		// in "passthru", we query specific fds first.
#if 0
	printf("pup1=%d\n",pup[1]);
	printf("pdn0=%d\n",pdn[0]);
	printf("per1=%d\n",per[1]);
#endif
		close(pup[1]);
		close(pdn[0]);
		close(per[1]);

		debug("pup0=%d\n",pup[0]);
		debug("pdn1=%d\n",pdn[1]);
		debug("per0=%d\n",per[0]);
#if 1
		if (pup[0]!=1) {
			dup2(pup[0],1);
			close(pup[0]);
			pup[0]=1;
		}
		if (pdn[1]!=2) {
			dup2(pdn[1],2);
			close(pdn[1]);
			pdn[1]=2;
		}
		if (per[0]!=3) {
			dup2(per[0],3);
			close(per[0]);
			per[0]=3;
		}
#endif
	}
	debug("execv=(%d %d) (%d %d) (%d %d)\n",pup[0],pup[1],pdn[0],pdn[1],per[0],per[1]);
#endif
}

void create_pipes(char *cmd,char **pts,int *ptm,int *pup,int *pdn,int *per) {

	debug("> %s\n",__FUNCTION__);

        /* no command specified or pty forced by "-t":
         * use a pseudo terminal:
         */

        if (strcmp(cmd,"")==0)
                do_pty=1;
        debug("do_pty=%d\n",do_pty);
        if (do_pty) {
                if ((*pts=getpty(ptm,O_RDWR,0622,0))==NULL)
                        die("getpty");
                debug("pseudo terminal on %s\n",pts);
        }
        else {
                /* if we don't have a pty-pipe, we have to create a bidir-pipe
                 * (update: a triple-pipe: stdin, stdout, stderr)
                 * because we have to talk to the child-process and gsswrap()
                 * all its messages. directly sending data to the socket is
                 * not allowed for the child process.
                 *
                 * - nb ... it's really a triple-pipe, including stderr.
                 */

                if (pipe(pup)==-1)              // read
                        die("pipe");
                if (pipe(pdn)==-1)              // write
                        die("pipe");
                if (pipe(per)==-1)              // read error
                        die("pipe");
                /* 
                 * uplink is for: parent <-- child (stdout of child)
                 * downlk is for: parent --> child (stdin of child)
                 * error  is for: parent <-- child (stderr of child)
                 */

                debug("pup0=%d pup1=%d pdn0=%d pdn1=%d per0=%d per1=%d\n",
                        pup[0],pup[1],
                        pdn[0],pdn[1],
                        per[0],per[1]);
	}


#if 0
        memset(&sa,0,sizeof(sa));
        //sa.sa_handler=sigchld;
        sa.sa_handler=SIG_IGN;
        //sa.sa_flags=SA_NOCLDWAIT|SA_NOCLDSTOP|SA_RESTART|SA_SIGINFO;
        //sa.sa_flags=SA_NOCLDSTOP|SA_RESTART|SA_SIGINFO;
        //sa.sa_flags=SA_RESTART|SA_NOCLDWAIT|SA_NOCLDSTOP;
        /* 
         * we want: NOCLDSTOP since we aren't interested in tracing the children
         * we do not want NOCLDWAIT (== we want CLDWAIT), since we'd like to have
         * our children zombied, so that we get all their data from the pipes until
         * we call waitpid()
         */
        sa.sa_flags=SA_RESTART|SA_NOCLDSTOP;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGCHLD,&sa,NULL);

#endif
}

void create_x11(int *xs) {
char *fqdn;
        if ((*xs=x11_create_listener())==-1)
                die("x11_create_listener");
        /* determine display location */
#if 1
        fqdn=dnslookup(uts.nodename);
        debug("> fqdn=%s\n",fqdn);
        if ((xdisploc=(char *)malloc(strlen(fqdn)+1+5+1))==NULL)
                die("out of memory");
        sprintf(xdisploc,"%s:%d",fqdn,xdispno);
        free(fqdn);
#endif
        //if ((xdisploc=(char *)malloc(9/*localhost*/+1+5+1))==NULL)
        //      die("out of memory");
        //sprintf(xdisploc,"localhost:%d",xdispno);
        // potential race condition
        //send_dsploc(as);

        if ((xunixloc=(char *)malloc(SYS_NMLN+6+5+1))==NULL)
                die("out of memory");
        sprintf(xunixloc,"%s/unix:%d",uts.nodename,xdispno);
}

int remote(int s,char *cmd,int nlf,int nrf,char **lfwd_list,char **rfwd_list) {
char *pts=NULL;
int ptm;
fwd_desc_entry_t *fwde;
int *fwdfd;
int xs=-1;
int pup[2]={-1,-1},pdn[2]={-1,-1},per[2]={-1,-1};
initcon_t *is;

	debug("> remote cmd=[%s] do_x11fwd=%d\n",cmd,do_x11fwd);
	if (s)
		dup2(s,0);
	// XXX 
	// closeall_but(0,-1);	// fd==0

	if (read(0,&eflag,4)!=4)
		die("couldn't read endian signature");
	eflag=ntohl(eflag);

	debug("got eflag=%.8x\n",eflag);

        //if (check_proto_version(s)==-1)
	//	die("wrong GSH-protocol version");
	// write_version(0);

	if (gssaccept(0)==-1)
		die("gssaccept failed");
        /* once we are done with gssaccept,
	 * we dont need the proxy certificate
	 * any more. remove it now!
	 */

	//is=malloc(sizeof(initcon_t));
	//if (read(as,is,sizeof(initcon_t))!=sizeof(initcon_t))

	// if (read_token(as,(void *)&is,&i)!=sizeof(initcon_t))
	// 	die("read init struct");

	// read exactly one encrypted packet.

	gss_buffer_desc wbuf;
	gss_set_encryption(1);		/* force encryption */
	// gssread(0,&wbuf,&more);
	gssread(0,&wbuf,NULL);
	gss_set_encryption(do_encrypt);	/* restore encryption */
	is=wbuf.value;

	//debug("reading initcon more=%d\n",more);

	/* check if our secret matches the secret of
	 * the other glogin process.
	 */
	// printf("CHECK SECRET>\n");
	// int i;
	// for(i=0;i<16;i++) printf("%.2x ",g_ic.secret[i]);
	// printf("\n");

#ifndef I2G
	if (memcmp(g_ic.secret,is->secret,16)) {
	struct sockaddr_in s_in;
	socklen_t len;
		len=sizeof(s_in);
		if (getpeername(0,(struct sockaddr *)&s_in,&len)==-1)
			memset(&s_in,0,sizeof(s_in));
		logger(LOG_AUTH,"hijack attempt from %s %s",
			inet_ntoa(s_in.sin_addr),cert_subject);
		exit(-1);
	}
#endif

	debug("session accepted");

	/* we close all fds again ... globus forgets to
	 * close one fd and we'll see an open ref to /dev/null
	 * that's ugly, so we fix that now.
	 */ 
	closeall_but(0,-1);

	/* now we definitely *ONLY* have fd0 which is our
	 * socket for tunneling
	 */

	/* create the x11 listener now
	 * because we need the DISPLAY
	 * variable in the shell
	 */
	if (!do_nocmd) create_pipes(cmd,&pts,&ptm,pup,pdn,per);
	if (do_x11fwd) create_x11(&xs);
	debug("x11=%d\n",xs);
	//if (!do_nocmd) create_pipes(cmd,&pts,&ptm,pup,pdn,per);

	// create the forwarders _before_ executing the commands.
	fwd_create_listeners(&fwdfd,nrf,do_remote,do_x11fwd,xs,rfwd_list);
	fwd_create_targets(&fwde,nlf,do_remote,do_x11fwd,lfwd_list);

	if (!do_nocmd) execv_cmd(cmd,is,pts,ptm,pup,pdn,per);
	debug("execv=(%d %d) (%d %d) (%d %d)\n",pup[0],pup[1],pdn[0],pdn[1],per[0],per[1]);
	free(is);

	// and _not_ here, or else the command is already running w/o
	// forwarders, but might need them!
	//fwd_create_listeners(&fwdfd,nrf,do_remote,do_x11fwd,xs,rfwd_list);
	//fwd_create_targets(&fwde,nlf,do_remote,do_x11fwd,lfwd_list);
	passthru(pup[0],pdn[1],per[0],0,fwdfd,fwde);

#if 0
	if (pup[1]>=0) close(pup[1]);
	if (pdn[1]>=0) close(pdn[0]);
	if (per[1]>=0) close(per[1]);
	close(as);
	if (xs>=0) close(xs);
#endif
	//closeall_but(-1);	/* hm? ... close all? */
	closeall_but(0,-1);	/* hm? ... close all? */
	//kill(-cmdpid,SIGKILL);
	//waitpid(cmdpid,&xs,0);	/* we need no steenking status */
	debug("child with pid %d exited with status %d\n",cmdpid,xs);
	return 0;
}

#ifdef DEBUG
void logargs(int argc,char *argv[]) {
char buf[1024];
int i;
	sprintf(buf,"%d:",argc);
	for(i=0;i<argc;i++) {
		strncat(buf," ",sizeof(buf));
		strncat(buf,argv[i],sizeof(buf));
		debug("arg%d=%s\n",i,argv[i]);
	}
	debug("%s\n",buf);
}

void fd2syslog(int fd) {
char buf[1024];
int p[2];
int i;

	if (pipe(p)==-1)
		die("pipe");
	switch(fork()) {
	case -1:
		die("fork");
	case 0:
		for(;;) {
			i=read(p[0],buf,sizeof(buf));
			if (i<1) die("fd2syslog: read");
			buf[i]=0;
			syslog(LOG_DAEMON|LOG_NOTICE,"%s\n",buf);
		}
		break;
	default:
		dup2(p[1],fd);
		close(p[0]);
		break;
	}
}
#endif

void send_initseq() {
initseq_t sq;
char *range;
int i;

	//debug("> send_initseq\n");

	/*
	 * this is an ugly workaround:
	 *
	 * if we want to tell the peer where we listen at,
	 * we have to fork() and exit(), to make the gass-
	 * cache empty itself.
	 *
	 * Mon Jun 28 18:43:50 CEST 2004 herp
	 * - transmit ipaddr
	 * - transmit endian signature
	 * - transmit secret key
	 * Sat Feb 19 10:41:12 CET 2005 herp
	 * - transmit initial termios.c_cc
	 * - transmit initial winsize
	 */


	debug("opt_port now %d\n",opt_port);

	if (read_rand(g_ic.secret,16)==-1)
		die("read_rand");
	/*
	 * Lo and behold! If we just exit, the user-proxy
	 * will be deleted too BECAUSE IT IS STORED IN THE
	 * GASS CACHE! it's only a stupid "accident" if the
	 * child process is fast enough!
	 *
	 * -- this is what is called a "race condition".
	 */

	// XXX GT4
	copy_x509_user_proxy(&fname_x509_up);
#if 1
	debug("forking\n");
	switch(fork()) {
	case -1:
		die("fork");
	case 0:
		setsid();		// neccessary to keep lcgpbs from sending SIGTERM/SIGKILL to us
		self_pid=getpid();	// this is the child process
		debug("pid after fork=%d\n",self_pid);
		return;
	default:
		break;
	}
#endif
	/*
	 * send the initialisation sequence
	 * from glogin" to glogin
	 */

	/* send port, ipaddr, secret key, version length, some reserveds ... */
	sq.sq_port=htons(opt_port);
	//sq.sq_inaddr=*(struct in_addr *)self_hp->h_addr_list[0];
	sq.sq_inaddr=self_addr;
	sq.sq_beef=0xfeedbeef;
	sq.sq_vers_len=strlen(GSH_PROTOCOL);
	memcpy(sq.sq_secret,g_ic.secret,16);
	write(1,&sq,sizeof(sq));
	i=strlen(GSH_PROTOCOL)+1;
	range=alloca(i);
	strcpy(range,GSH_PROTOCOL);
	debug("sending %d bytes = %s\n",i,range);
	write(1,range,i);

	// logger(LOG_ERR,"sent sequence to stdout, exiting\n");
	debug("sent sequence to stdout, exiting\n");
	exit(0);
}

void flag_winsize(int nsig) {
// XXX USE MUTEX TO LOCK ???
// event could probably be lost
// due to race conditions
	do_winsize=1;
}

int main(int argc,char *argv[]) {
struct sigaction  sa;
char *host=NULL,*exec=NULL;
char *cmd=NULL;
char *cp;
int nlf=0,nrf=0;	/* number of local, number of remote forwarders */
fwd_desc_entry_t *fwde;	/* forwarder target table */
int *fwdfd;		/* array of forwarder fd to listen to */
char **lfwd_list=NULL;	/* list containing local forwarders, parsed from argv */
char **rfwd_list=NULL;	/* list containing remote forwarders, parsed from argv */
gss_buffer_desc wbuf;
int i;
int s,sc,sl;

#ifdef DEBUG
	logargs(argc,argv);
#endif
	//fd2syslog(2);
	self=basename(argv[0]);
	self_pid=getpid();

	while((i=getopt(argc,argv,"fvgNCc:he:rp:l:L:R:i:o:q:s:txX"))!=EOF) {
		switch(i) {
		case 'v':
			printf("GSH-%s %s\n",
				VERSION,GSH_PROTOCOL);
			exit(0);
		case 'f':
			do_fork=1;
			break;
		case 'c':
			if (cmd) usage();	/* only one command please */
			cmd=strdup(optarg);
			debug("cmd=[%s]\n",cmd);
			break;
		case 'C':
			do_encrypt=0;
			break;
		case 'N':
			do_nocmd=1;
			break;
		case 'g':
			do_staging=1;
			break;
		case 'h':
			usage();
		case 'e':
			opt_errsock=strdup(optarg);
                	if (opt_errsock) {
			struct sockaddr_in s_in;
                        	if (resolv(opt_errsock,&s_in.sin_addr)==-1)
                                	die("resolv");
                        	if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
                                	die("socket");
                        	s_in.sin_family=AF_INET;
                        	s_in.sin_port=htons(4041);
                        	if (connect(s,(struct sockaddr *)&s_in,sizeof(s_in))==-1)
                                	die("connect");
                        	dup2(s,2);
				uname(&uts);
                        	fprintf(stderr,"nodename=%s pid=%d\r\n",uts.nodename,getpid());
                        	fflush(stderr);
                	}
			break;
		case 'r':
			do_remote=1;
			break;
		case 'p':
			opt_port=atoi(optarg);
			if (opt_port>0xffff)
				die("illegal port");
			if ((cp=index(optarg,':'))) {
				opt_addr=strdup(cp+1);
				do_addr=1;
			}
			break;
		case 'l':
			do_exec=1;
			exec=strdup(optarg);
			break;
		case 'i':
			opt_infile=strdup(optarg);
			break;
		case 'o':
			opt_outfile=strdup(optarg);
			break;
		case 'q':
			opt_queue=strdup(optarg);
			break;
		case 's':
			opt_rsl=strdup(optarg);
			break;
		case 't':
			do_pty=1;
			break;
		case 'x':
			do_x11fwd=0;
			break;
		case 'X':
			do_x11fwd=1;
			break;
		case 'L':
			//printf("local fwd optarg=%s\n",optarg);
			nlf++;
			lfwd_list=realloc(lfwd_list,nlf*sizeof(char *));
			lfwd_list[nlf-1]=strdup(optarg);
			debug("list:\n");
			for(i=0;i<nlf;i++)
				debug("%d %s\n",i,lfwd_list[i]);
			break;
		case 'R':
			nrf++;
			rfwd_list=realloc(rfwd_list,nrf*sizeof(char *));
			rfwd_list[nrf-1]=strdup(optarg);
			break;
		}
	}
	// addr not allowed in non-remote and required in remote
	if ((do_remote && do_addr==0) || (do_remote==0 && do_addr))
		usage();

	if (uname(&uts)==-1)
		die("uname");

	gss_set_encryption(do_encrypt);
#if 0
	if (optind+1<argc && cmd)	// -c and extra arguments
		usage();
#endif
#if 1
	host=argv[optind];
	//fprintf(stderr,"host=%s\n",host);
	// if no explicit -c switch, try to construct
	// it from the command line similar to ssh
	if (!cmd) {
        	i=optind+1;
        	if (i<argc) {
		char **args;
        	int j=i,n=0;
                	//fprintf(stderr,"optind=%d argc=%d\n",optind,argc);
			/* construct one single -c arg
			 * by copying all commands together
			 */
                	while(i<argc)
                        	n+=strlen(argv[i++]);
			n+=argc-optind-1;
                	// fprintf(stderr,"n=%d\n",n);
                	if ((cmd=(char *)malloc(n))==NULL)
                        	die("out of memory");
			//n=j;
                	strcpy(cmd,argv[j++]);
                	i=strlen(cmd);
                	while(j<argc) {		/* paste -c arg */
                        	cmd[i++]=' ';
                        	strcpy(cmd+i,argv[j]);
                        	i+=strlen(argv[j]);
                        	j++;
                	}
			//fprintf(stderr,"cmd=%s n=%d\n",cmd,n);
			if ((args=malloc((argc+3)*sizeof(char *)))==NULL)
				die("out of memory");
			//memcpy(args,argv,n*sizeof(char *));
			memcpy(args,argv,optind*sizeof(char *));
			argc=optind;
			args[argc++]="-c";
			args[argc++]=cmd;
			optind=argc;
			args[argc++]=host;
			args[argc]=NULL;
			argv=args;
#if 0
			fprintf(stderr,"argc=%d\n",argc);
			for(i=0;i<argc;i++)
				fprintf(stderr,"args%d=%s\n",i,args[i]);
#endif
		}
        }
#endif

	if (!cmd) {
		cmd="";
		do_pty=1;
	}
	/*
	 * ==================================
	 * code for remote instance of glogin
	 * ==================================
	 */
	if (do_remote) {
#ifdef I2G
        struct sockaddr_in s_in;
#endif
	struct in_addr addr;
		debug("do_remote\n");

#if 0
		if (argc-optind) { 
			debug("opt_remote: argc=%d optind=%d\n",argc,optind);
			//while(optind<argc)
			//	fprintf(stderr,"%s\n",argv[optind++]);
			//usage();
		}
#endif
		if ((self_hp=gethostbyname(uts.nodename))==NULL)
			die("gethostbyname"); 
		if (do_x11fwd==-1)
			do_x11fwd=1;
#if 0
		if (opt_errsock) {
			if (resolv(opt_errsock,&s_in.sin_addr)==-1)
				die("resolv");
			if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
				die("socket");
			s_in.sin_family=AF_INET;
			s_in.sin_port=htons(4041);
			if (connect(s,(struct sockadddr *)&s_in,sizeof(s_in))==-1)
				die("connect");
			dup2(s,2);
			fprintf(stderr,"nodename=%s\r\n",uts.nodename);
			fflush(stderr);
		}
#endif
		// so silly! in_addr_t != typedef struct in_addr
		addr.s_addr=inet_addr(opt_addr);
#ifdef I2G
                if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
                        die("socket");
                s_in.sin_family=AF_INET;
                s_in.sin_addr=addr;
                s_in.sin_port=htons(opt_port);
                if (connect(s,(struct sockaddr *)&s_in,sizeof(s_in))==-1)
                        die("connect");
#else
		sc=connect_nb(addr,opt_port);
		sl=create_listener(0);
		send_initseq();		// + force empty gass cache; this will do the fork-thing
		closeall_but(sc,sl,-1);	// neccessary for OpenSSH, else job is "running"
		s=tun_establish(sc,sl);
#endif
		remote(s,cmd,nlf,nrf,lfwd_list,rfwd_list);
		cleanup();
		debug("exiting...\n");
		// exit(0);
		_exit(0);
	}

	/*
	 * =================================
	 * code for local instance of glogin
	 * =================================
	 */

	if (cmd && argc-optind>1)
		die("cant process both -c and command list");
#ifndef I2G
	if (argc==optind)
		usage();
#endif

	if (opt_queue) {
	char *cp;
		cp=opt_rsl;
		/* (queue=%s) */
		opt_rsl=realloc(opt_rsl,(opt_rsl?strlen(opt_rsl):0)+strlen(opt_queue)+9+128);
		if (cp)
			cp+=strlen(cp);
		else
			*opt_rsl=0;
		strcat(opt_rsl,"(queue=");
		strcat(opt_rsl,opt_queue);
		strcat(opt_rsl,")");
	}
	// fprintf(stderr,"host=%s\n",host);
	// host=argv[optind];

	if (exec==NULL)
		//exec="/home/gup/hr/src/glogin";
		//exec="/home/globus/bin/glogin";
		//exec="/home/gup/hr/src/GLOGIN/glogin.sh";
		//exec="/opt/cg/bin/glogin";
		exec="bin/glogin";

	globus_module_activate(GLOBUS_COMMON_MODULE);
#ifndef WSGRAM
	globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE);
#endif
	globus_libc_gethostname(ghostname,sizeof(ghostname));
	debug("ghostname=%s\n",ghostname);
	if ((self_hp=gethostbyname(ghostname))==NULL)
		die("gethostbyname"); 
	self_addr=*(struct in_addr *)self_hp->h_addr_list[0];

	if (do_x11fwd==-1)	// undefined? -> default=yes
		do_x11fwd=1;
	if (do_x11fwd) {
		i=x11_init_display();
		//if (x11_init_display()==-1) {	// failed, so turn it off
		if (i==-1) {
		char **arg;
                        //fprintf(stderr,"X11 forwarding disabled since no display was found\n");
			// in I2G, we don't have a hostname on the end, so
			// just reusing it for setting "-x" wont work.
			// since argv is on the stack and not the heap,
			// realloc wont work either. so, make a copy and
			// append "-x".
			argc++;
			if ((arg=malloc((argc+1)*sizeof(char *)))==NULL)
				die("malloc");
			for(i=0;argv[i];i++)
				arg[i]=argv[i];
			arg[i++]="-x";		/* disable */
			arg[i]=NULL;
			argv=arg;
			do_x11fwd=0;
		}
		else x11_init_cookie();
	}
	// printf("pid=%d\n",getpid());
	// printf("host=%s glogin=%s cmd=%s\n",host,exec,cmd);
	debug("host=%s glogin=%s cmd=%s\n",host,exec,cmd);

	sl=create_listener(opt_port);
#ifdef I2G
        printf("-p %d:%s\n",opt_port,inet_ntoa(self_addr));
        fflush(stdout);
	s=accept_once(sl);
#else
	// fprintf(stderr,"gssexec(%s,%s,%d,%s)\n",
	// 	host,exec,opt_port,opt_rsl);
	gssexec(host,exec,opt_port,opt_rsl,argc,argv);
	if (opt_port==0)
		die("zero port");
	// if (is_private(r_addr))
	// 	die("private target IP addr");
	sc=-1;
	if ((s=maybe_accept(sl))!=-1)	{	// connection grid->WS established
		close(sl);
		sl=s;
	}
	else	// no connection was made from grid to WS
		sc=connect_nb(r_addr,opt_port);
	s=tun_establish(sc,sl);
#endif

	//debug("Connected to %s:%d. (socket=%d)\n",host,opt_port,s);
	// fprintf(stderr,"Connected to %s:%d. (socket=%d)\n",host,opt_port,s);
	// fflush(stderr);

	/* we already know if the endianness matches (sq.sq_beef)
	 * send the 'eflag' to the other side, so they also know
	 * what we know.
	 */
	i=htonl(eflag);
	if (write(s,&i,4)!=4)
		die("couldn't send endian flag");
	debug("sent eflag=%d\n",eflag);

	// write_version(s);
	// if (check_proto_version(s)==-1)
	// 	die("wrong GSH-protocol version");

	gssinit(s);

        // gss_display_name(&min,client_name,&name_buffer,NULL);
        // logger(LOG_INFO,"login %s\n",(char *)name_buffer.value);

#ifndef WSGRAM
	globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);	/* not needed anymore */
#endif
	globus_module_deactivate(GLOBUS_COMMON_MODULE);		/* not needed anyore */

	/* transmit the secret back again. if they match, all is fine */
	/* also send initial c_cc and winsize */
	if (ioctl(0,TIOCGWINSZ,&g_ic.ws)==-1)
		memset(&g_ic.ws,0,sizeof(g_ic.ws));
	if (ioctl(0,TCGETS,&g_tc)==-1)
		memset(&g_tc,0,sizeof(g_tc));
	memcpy(g_ic.c_cc,g_tc.c_cc,NCC);

	// logger(LOG_INFO,"sending secret back to remote\n");
	wbuf.length=sizeof(g_ic);
	wbuf.value=&g_ic;

	gss_set_encryption(1);		/* force encryption */
	gsswrite(s,&wbuf);
	gss_set_encryption(do_encrypt);	/* restore encryption */

	/* install window-size handler */
	sa.sa_handler=flag_winsize;
	sa.sa_flags=SA_RESTART;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGWINCH,&sa,NULL);

	if (opt_outfile) {
		if ((i=open(opt_outfile,O_WRONLY|O_TRUNC|O_CREAT,0600))==-1)
			die("open");
		if (dup2(i,1)==-1)
			die("dup2");
		close(i);
	}
        if (opt_infile) {
                if ((i=open(opt_infile,O_RDONLY|O_NONBLOCK))==-1)
                        die("open");
                if (dup2(i,0)==-1)
                        die("dup2");
                close(i);
        }
	/* globus forgets to close a filedescriptor, see also:
	 * globus_forgotten_fd.c (written by myself)
	 */

	/* if we don't have a command, we only need a socketfd in fd=0
	 * and no ptys. but on the other hand, error messages would be
	 * lost then. so, we still keep fd0,1,2. ssh does that anyway, too,
	 * but this leads, to not being able to log out when logged in
	 * via openssh. oh my.
	 */
	if (s!=3) {
		if (dup2(s,3)!=3)
			die("dup2");
		s=3;
	}
	closeall_but(0,1,2,s,-1);

	//i=16*1024;
	//i=4*(4096+5+32);
	//i=4*(32768+5+32);
	//i=4*(65536+5+32);
	// wert experimentell ermittelt:
	// ./gss-encryption-speed -b 16384 -k 10240 -c c -o /dev/null
	// [pid 26568] write(6, "\27\3\0@ Z\10V\v\7\376\236\324f\371\204\346H\214\335\201"..., 16421) = 16421
	//i=GSH_TOKEN_SIZE;
	//i=32768;
	//i=65536;
	i=128*1024;
	if (setsockopt(s,SOL_SOCKET,SO_SNDBUF,&i,sizeof(i))==-1)
		die("setsockopt");
	if (setsockopt(s,SOL_SOCKET,SO_RCVBUF,&i,sizeof(i))==-1)
		die("setsockopt");

	/* create forwarders */
	fwd_create_listeners(&fwdfd,nlf,do_remote,0,-1,lfwd_list);
	fwd_create_targets(&fwde,nrf,do_remote,do_x11fwd,rfwd_list);

#if 0
	if (!do_fork)
		i=0;	// ttyfd for passthru() (see below)
	else {
		i=-1;	// do not listen to ttyfd when in background!
		switch(fork()) {
		case -1:
			die("fork");
		case 0:
			setsid();
			break;
		default:
			_exit(0);
		}
	}
#endif
	if (do_fork) {
		switch(fork()) {
		case -1:
			die("fork");
		case 0:
			setsid();
			break;
		default:
			_exit(0);
		}
	}

	if (do_pty && !do_fork)
		tty_makeraw(0);
	//passthru(i,1,2,s,fwdfd,fwde);
	passthru(do_fork?-1:0,1,2,s,fwdfd,fwde);
	if (do_pty && !do_fork)
		tty_restore(0);
	cleanup();

	/* informative message only if we had a shell session */
	if (cmd[0]==0)
		fprintf(stderr,"Connection closed by foreign host.\n");
	debug("exiting...\n");
	exit(0);
}
