
#ifndef __glogin_h_
#define __glogin_h_

#ifdef __linux__
#define	PATH_XAUTH		"/usr/X11R6/bin/xauth"
#define	HAVE_GETPT
#define	HAVE_UNSETENV
#define	HAVE_BASENAME
#endif

#ifdef __sun__
#define PATH_XAUTH              "/usr/X/bin/xauth"
#define HAVE_GETPTMX
#define HAVE_STROPTS
#endif

#define	GSH_PROTOCOL	"1.1"
#define	GSH_TOKEN_SIZE	16421
// #ifdef	CROSSGRID
// #define	VERSION		"1.1.3-cg"
// #else
// #define	VERSION		"1.1.3 (ws)"
// #endif

// prototypes

#include	<sys/types.h>
#include        <sys/ioctl.h>
#include        <sys/termios.h>
#include	<netinet/in.h>

#include	<globus_gass_server_ez.h>
#include        <globus_gram_client.h>
#include        <globus_gram_protocol.h>

/* this structure is sent to glogin on the
 * workstation. (followed by the version string,
 * which is sent seperately, due to alignment
 * issues)
 */
typedef struct {
        uint32_t sq_beef;
        struct in_addr sq_inaddr;
	u_int sq_reserved1;
        u_short sq_port;
	u_short sq_reserved2;
        char sq_secret[16];
	char sq_vers_len;
	char sq_reserved3[3];
} __attribute__((packed)) initseq_t;

typedef struct {
	struct winsize ws;
	char c_cc[NCC];
	char secret[16];
} initcon_t;

typedef struct {
        char *cmd;
        char **args;
        int nargs;
        struct passwd *pw;
} rcmd_t;


void usage(void);

void tty_restore(int);
void tty_makeraw(int) ;

int read_rand(void *,int);
int close_rand(void);
void bin2hex(void *,void *,int);

#define warn(s)  real_die(0,__FILE__,__LINE__,s)
#define die(s)  real_die(1,__FILE__,__LINE__,s)
void real_die(int,char *,int,char *);

int resolv(char *,struct in_addr *);
char *dnslookup(char *);

#ifndef HAVE_UNSETENV
int unsetenv(const char *);
#endif

extern struct in_addr r_addr;
extern initcon_t g_ic;
extern int eflag;
uint16_t swap16(uint16_t);
uint32_t swap32(uint32_t);
void swapws(struct winsize *);

int save_read(int,void *,int);
// int read_token(int,void **,uint32_t *,int *);

int read_token_enc(int,void **,uint32_t *,int *);
int read_token_clear(int,void **,uint32_t *,int *);
int read_token_unbuffered(int,void *,uint32_t);
int (*read_token)(int,void **,uint32_t *,int *);

int write_token_enc(int,void *,uint32_t);
int write_token_clear(int,void *,uint32_t);

int (*write_token)(int,void *,uint32_t);

void logger(int,char *,...);
void xdump(char *,void *,int);
#ifdef DEBUG
void debug(char *,...);
#else
#define debug(fmt...)
#endif

int is_private(struct in_addr);
void copy_x509_user_proxy(char **);

int maxfd(int,fd_set *);
void set_timeout(int);

void cleanup();

char *basename(char *);
char *getpty(int *,int,mode_t,int);

void closeall_but(int,...);

rcmd_t *rcmd_split(char *,int);

int create_listener(u_short);
int connect_nb(struct in_addr,u_short);
int maybe_accept(int);
int accept_once(int);
int tun_establish(int,int);

// from gss_glue

void gss_set_encryption(int);
int gsswrite(int,gss_buffer_desc *);
int gssread(int,gss_buffer_desc *,int *);
int gssread_is(initcon_t **);
int gssaccept(int);
int gssinit(int);
int gssexec(char *,char *,u_short,char *,int,char *[]);
void gsserror(OM_uint32,OM_uint32);

// XXX send_
void send_error_msg(int,char *);

#endif


