/*
 * Tue Oct 28 13:57:03 CET 2003 herp - Herbert Rosmanith
 *
 */

#include        <stdio.h>
#include        <stdlib.h>
#include        <stdarg.h>
#include	<sys/types.h>
#include	<sys/ioctl.h>
#include	<sys/stat.h>
#include	<sys/socket.h>
#include	<sys/syslog.h>
#include	<sys/termios.h>
#include	<signal.h>
#include	<string.h>
#include	<strings.h>
#include	<fcntl.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<netdb.h>
#include	<unistd.h>
#include	<errno.h>

#include	"glogin.h"
#include	"channel.h"
#include	"x11.h"

extern struct termios g_tc;
extern pid_t self_pid;
extern char *opt_outfile;
extern int opt_port;
extern int do_restore;
extern int do_remote;
extern int enc_mode;

#ifdef PROXY_DELEGATION
extern char x509_user_proxy[];
extern time_t ctime_user_proxy;
#endif

static int rand_fd=-1;

int read_rand(void *p,int len) {
register char *cp=(char *)p;

	if (rand_fd==-1)
		if ((rand_fd=open("/dev/urandom",O_RDONLY))==-1)
			return -1;
	return read(rand_fd,cp,len);
}

int close_rand() {
	if (rand_fd==-1)
		return -1;
	close(rand_fd);
	rand_fd=-1;
	return 0;
}

int cat_file(char *name,int fd) {
char buf[4096];
int fd_in;
int i;
	if ((fd_in=open(name,O_RDONLY))==-1)
		return -1;
	for(;;) {
		if ((i=read(fd_in,buf,sizeof(buf)))<1)
			return i;
		write(fd,buf,i);
		if (i!=sizeof(buf))
			break;
	}
	return 0;
}

void usage() {
	if (!do_remote) {
	char *fname;
		if ((fname=alloca(256))) {
			snprintf(fname,256,"/usr/share/doc/glogin-%s/GUP-OSS-License.txt",VERSION);
			if (cat_file(fname,2)==-1)
				fprintf(stderr,"\nThe license could not be displayed. Please go to the URL mentioned below.\n\n");
		}
	}
	fprintf(stderr,"(c)GUP, Joh. Kepler University Linz, Austria\n\
glogin [-hrtvfgxXCN] [-l binary] [-o file] [-p port] [-s rsl] [-i file] [-o file] [-c cmd] [-L port:host:port...] [-R port:host:port...] host\n\
Please go to http://www.gup.uni-linz.ac.at/glogin for more details\n");
	exit(-1);
}

/* this version saves about 40 bytes of code
 * compared to the version below.
 */

void bin2hex(void *dst,void *src,int len) {
static char xtab[]="0123456789abcdef";
register u_char *cd=(u_char *)dst;
register u_char *cs=(u_char *)src;
int i;
        for(i=0;i<len;i++) {
		*cd++=xtab[*cs>>4];
		*cd++=xtab[*cs&15];
		cs++;
	}
	*cd=0;
}

#if 0
void bin2hex(void *dst,void *src,int len) {
register u_char *cd=(u_char *)dst;
register u_char *cs=(u_char *)src;
register u_char ch;
int i;

        for(i=0;i<len;i++) {
                ch=*cs>>4;
		*cd=ch+(ch<10?'0':'a'-10);
                cd++;
                ch=*cs&0xf;
		*cd=ch+(ch<10?'0':'a'-10);
                cd++;
                cs++;
        }
        *cd=0;
}
#endif

/* not needed anymore */

#if 0
void hex2bin(void *dst,void *src,int len) {
register u_char *cd=(u_char *)dst;
register u_char *cs=(u_char *)src;
register u_char ch;
int i;

	for(i=0;i<len;i++) {
		ch=*cs-(*cs<='9'?'0':'a'-10);
		ch<<=4;
		cs++;
		ch|=*cs-(*cs<='9'?'0':'a'-10);
		*cd=ch;
		cs++;
		cd++;
	}
}
#endif

uint16_t swap16(uint16_t sh) {
	return (sh>>8)|(sh<<8);
}

uint32_t swap32(uint32_t l) {
	return swap16(l>>16)|(swap16(l)<<16);
}

void swapws(struct winsize *ws) {
	ws->ws_row=swap16(ws->ws_row);
	ws->ws_col=swap16(ws->ws_col);
	ws->ws_xpixel=swap16(ws->ws_xpixel);
	ws->ws_ypixel=swap16(ws->ws_ypixel);
}

int resolv(char *host,struct in_addr *addr) {
struct hostent *hp;

        if (inet_aton(host,addr))
                return 0;
        if ((hp=gethostbyname(host))==NULL)
                return -1;
        *addr=*(struct in_addr *)*hp->h_addr_list;
        return 0;
}

char *dnslookup(char *host) {
struct hostent *hp;
	if ((hp=gethostbyname(host))==NULL)
		return NULL;	
	return strdup(hp->h_name);
}

void real_die(int do_exit,char *file,int lineno,char *s) {
static int in_die=0;
char buf[1024];

        snprintf(buf,sizeof(buf),"%s%s %d: %s",
		do_remote?"remote ":"",file,lineno,s?s:"-");
	if (in_die) {
        	logger(LOG_ERR,"double fault: %s",buf);
		return;
	}
	in_die=1;
        if (errno) {
                strncat(buf,": ",sizeof(buf));
                strncat(buf,strerror(errno),sizeof(buf));
        }
        strncat(buf,"\r\n",sizeof(buf));
        logger(LOG_ERR,buf);
#ifndef I2G
	/* send error message to local glogin */
	if (do_remote) {
		 if (getpid()==self_pid)
			send_error_msg(0,buf);
		else
			write(2,buf,strlen(buf));
	}
	else
#endif
	{
		fprintf(stderr,buf);
		tty_restore(0);
	}
        if (do_exit)
		_exit(-1);
	in_die=0;
}

#ifndef HAVE_UNSETENV
extern char **environ;

int unsetenv(const char *name) {	/* shamelessly borrowed from glibc */
char **ep,**dp;
int n;

        ep=environ;
        n=strlen(name);
        while(*ep) {
                if (memcmp(*ep,name,n)==0 && (*ep)[n]=='=') {
                        /* found it */
                        dp=ep;
                        do {
                                dp[0]=dp[1];
                        } while(*++dp);
                }
                else
                        ep++;
        }
        return 0;
}
#endif

int is_private(struct in_addr addr) {

	addr.s_addr=ntohl(addr.s_addr);
	debug("is_private(addr=%.8x)\n",addr.s_addr);

	/* 10/8 */
	if (((addr.s_addr & 0xff000000)==0x0a000000) ||
	/* 172.16/12 */
	    ((addr.s_addr & 0xfff00000)==0xac100000) ||
	/* 192.168/16 */
	    ((addr.s_addr & 0xffff0000)==0xc0a80000))
		return 1;
	return 0;
}

/*
 * copy the ${X509_USER_PROXY} to /tmp/x509up_<tmpname>
 *
 * caller is responsible for freeing x509_up
 */

void copy_x509_user_proxy(char **x509_up) {
char tfnam[]="/tmp/x509up_XXXXXX";
char buf[4096];
char *cp;
int fd1,fd2;
int i;

	debug("copy_x509_user_proxy: X509_USER_PROXY=%s\n",getenv("X509_USER_PROXY"));
	if ((cp=getenv("X509_USER_PROXY"))==NULL)
		die("getenv");
	if ((fd1=open(cp,O_RDONLY))==-1)
		die("open");
	if ((fd2=mkstemp(tfnam))==-1)
		die("mkstemp");
	if (fchmod(fd2,0600)==-1)
		die("fchmod");
	/*X509_USER_PROXY=/tmp/x509up_XXXXXX.*/
	/*12345678901234567890123456789012345*/
	if ((cp=(char *)malloc(35))==NULL)
		die("out of memory");
	sprintf(cp,"X509_USER_PROXY=%s",tfnam);
	putenv(cp);
	/* copy file */
	for(;;) {
		if ((i=read(fd1,buf,sizeof(buf)))==-1)
			die("read");
		if (i==0)
			break;
		if (write(fd2,buf,i)!=i)
			die("write");
	}
	close(fd1);
	close(fd2);
	*x509_up=strdup(cp+16);	/* first char after '=' */

	// DONT free this, or else the
	// environment variable will go away!
	// free(cp);

	return;
}

void tty_makeraw(int fd) {
struct termios tm;
	// if (ioctl(fd,TCGETS,&g_tc)==-1)
	// 	die("ioctl TCGETS");
	memset(&tm,0,sizeof(tm));
	tm.c_cc[VMIN]=1;
	tm.c_iflag=IGNBRK|IGNPAR;
	tm.c_cflag=CREAD|CS8|CLOCAL|HUPCL|B115200;
	if (ioctl(fd,TCSETS,&tm)==-1)
		// die("ioctl");
		do_restore=0;
	else do_restore=1;
}

void tty_restore(int fd) {
	if (do_restore)
		ioctl(fd,TCSETS,&g_tc);
		// nobody wants to know ...
		// if (ioctl(fd,TCSETS,&g_tc)==-1)
		// 	die("ioctl TCSETS");
}

/* returns the biggest filedscriptor in a fd_set */

int maxfd(int n,fd_set *fds) {
int i;
        for(i=n;i>=0;i--) {
                if (FD_ISSET(i,fds))
                        return i;
        }
        return -1;
}

void timeout(int nsig) {
        logger(LOG_ERR,"timeout\n");
        exit(-1);
}

void set_timeout(int n) {
struct sigaction sa;
        sa.sa_handler=n?timeout:SIG_IGN;
        sa.sa_flags=0;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGALRM,&sa,NULL);
        alarm(n);
}

void cleanup() {
struct stat st;
#ifndef I2G
char *cp;
#endif

	/* cleanup: local glogin always, remote only
	 * the main process
	 */
	if (!do_remote || getpid()==self_pid) {
        	if (xdisploc) x11_remove_Xauthority(xdisploc);
        	if (xunixloc) x11_remove_Xauthority(xunixloc);
	};
	/* if no output has been produced, remove the output file
	 */
	if (!do_remote && opt_outfile!=NULL)
		if (fstat(1,&st)!=-1 && S_ISREG(st.st_mode) && st.st_size==0)
			unlink(opt_outfile);
#ifdef PROXY_DELEGATION
	if (do_remote) {
		/* we have to stat() again - some other process could
		 * have created a new proxy. in this case, we must not
		 * delete it. hm, is this process prone to race conditions?
		 * we should lock the file... (-> mandatory file locks)
		 */
		if (stat(x509_user_proxy,&st)==0 && st.st_ctime==ctime_user_proxy)
			unlink(x509_user_proxy);
	}
#endif
#ifndef I2G
	if (do_remote) {
        	if ((cp=getenv("X509_USER_PROXY")))
                	unlink(cp);
        	if ((cp=getenv("X509_USER_CERT")))
                	unlink(cp);
        	if ((cp=getenv("X509_USER_KEY")))
                	unlink(cp);
	}
#endif
}


/*
 * closing all fds is done this way:
 * first, a fd_set containing all fds not to
 * be closed is constructed. second, a fd_set
 * with all open fds is constructed from
 * /proc/self/fd. then, all fds present in
 * the second set and absent in the first
 * set are closed.
 *
 * this is better than closing during opendir(),
 * since, although unlikely, we could possibly close
 * our own dir-fd * by accident and there's no
 * way to get it from DIR * (opaque typedef).
 */

void closeall_but(int fd,...) {
fd_set xfds,sfds;
va_list ap;
struct dirent *de;
char *cp,*ep;
#ifdef __linux__
static char self_fd[]="/proc/self/fd";
#endif
#ifdef __sun__
char self_fd[15]; /* /proc/XXXXX/fd\0 */
#endif
DIR *d;

	debug("closeall_but ");
	FD_ZERO(&xfds);
	va_start(ap,fd);
	for(;;) {
		if (fd==-1)	/* end of parameter list */
			break;
		debug(" %d",fd);
		FD_SET(fd,&xfds);
		fd=va_arg(ap,int);
	}
	va_end(ap);
#ifdef DEBUG
	FD_SET(254,&xfds);
#endif

#ifdef __sun__
	sprintf(self_fd,"/proc/%d/fd",(int)self_pid);
#endif
	FD_ZERO(&sfds);
        if ((d=opendir(self_fd))==NULL) return;
        while((de=readdir(d))) {
                if (de->d_name[0]=='.')
                        continue;
                fd=strtol(de->d_name,&ep,10);
		if (*ep==0)	// we've got a number
			FD_SET(fd,&sfds);
        }
	closedir(d);

	cp=(char *)&xfds;
	ep=(char *)&sfds;
	for(fd=0;fd<sizeof(fd_set);fd++)
		ep[fd]&=cp[fd]^0xff;	// C = !A & B
	debug("closeall_but -  closing fds:");
	for(fd=0;fd<8*sizeof(fd_set);fd++)
		if (FD_ISSET(fd,&sfds)) {
			debug(" %d",fd);
			close(fd);
		}
		
}

#ifdef HAVE_BASENAME
extern char *basename(char *);
#else
char *basename(char *s) {
char *cp;
        if ((cp=rindex(s,'/'))==NULL) return s;
        return cp+1;
}
#endif

#ifdef HAVE_GETPTY
/*
 * this is a wrapper around _getpty() which
 * saves/restores the sigchld-handler
 */
char *getpty(int *filedes,int flags,mode_t mode,int nofork) {
struct sigaction sa,osa;
char *cp;

        if (nofork==0) {
                sa.sa_handler=SIG_DFL;
                sa.sa_flags=0;
                sigemptyset(&sa.sa_mask);
                sigaction(SIGCHLD,&sa,&osa);
        }
        cp=_getpty(filedes,flags,mode,nofork);
        if (nofork==0) sigaction(SIGCHLD,&osa,NULL);
        return cp;
}
#else
#if defined(HAVE_GETPT) || defined(HAVE_GETPTMX)
extern char *ptsname(int);      /* where is this??? */
int getpt();                    /* where is this??? */
int grantpt(int);               /* where is this??? */
int unlockpt(int);              /* where is this??? */

#ifdef HAVE_GETPTMX
inline int getpt() {
        return open("/dev/ptmx",O_RDWR|O_NOCTTY);
}
#endif

char *getpty(int *filedes,int flags,mode_t mode,int nofork) {
int ptm=-1;
char *cp;

        if ((ptm=getpt())==-1)
                goto error_return;
        if (grantpt(ptm)==-1)
                goto error_return;
        if (unlockpt(ptm)==-1)
                goto error_return;
        if ((cp=ptsname(ptm))==NULL)
                goto error_return;
        *filedes=ptm;
        return cp;
error_return:
        if (ptm>-1)
                close(ptm);
        return NULL;
}
#endif
#endif

/*
 * split a command in string form into pieces
 * usable to pass to execv()
 */

rcmd_t *rcmd_split(char *str,int login_shell) {
static rcmd_t rcmd;
static struct passwd pw;
char *cp,*arg;
char st;
int empty_command=1;
int i;
		
	debug("rcmd_split str=%s\n",str);

        rcmd.cmd=NULL;
        rcmd.args=NULL;
        cp=str;
        while(*cp==' ' || *cp=='\t') cp++;              /* skip while blank */
        if (*cp) empty_command=0;
        i=0;
        for(;;) {
                //arg=cp;
		/* if string is enclosed in ' or ", dont split when we see a blank */
		st=(*cp=='\'' || *cp=='"')?*cp:0;
		arg=st?++cp:cp;	/* skip the ' or " */
		debug("arg=%s\n",arg);
                //while(*cp!=' ' && *cp!='\t' && *cp) cp++;       /* skip until blank */
                while(((st && *cp!=st) || (st==0 && *cp!=' ' && *cp!='\t')) && *cp) cp++;       /* skip until blank */
		if (st || *cp)	/* clear the ' or " or <SP> */
                        *(cp++)=0;
                rcmd.args=(char **)realloc(rcmd.args,(i+1)*sizeof(char *));
                rcmd.args[i++]=arg;
                if (*cp==0)
                        break; 
                while(*cp==' ' || *cp=='\t') cp++;              /* skip while blank */
                if (*cp==0)
                        break;
        }
        rcmd.nargs=i;
        rcmd.args=(char **)realloc(rcmd.args,(rcmd.nargs+1)*sizeof(char *));
        if ((rcmd.pw=getpwuid(getuid()))==NULL) {
                rcmd.pw=&pw;
                pw.pw_shell="/bin/sh";
                pw.pw_dir="/";
        }
        if (empty_command)
                rcmd.args[0]=rcmd.pw->pw_shell;
        rcmd.cmd=strdup(rcmd.args[0]);
        if (login_shell) {
                cp=basename(rcmd.args[0]);
                rcmd.args[0]=(char *)malloc(strlen(cp)+2);
                strcpy(rcmd.args[0],"-");
                strcat(rcmd.args[0],cp);
        }
        else rcmd.args[0]=strdup(basename(rcmd.args[0]));
        rcmd.args[rcmd.nargs]=NULL;
	for(i=0;i<rcmd.nargs;i++)
		debug("arg%d=%s\n",i,rcmd.args[i]);
        return &rcmd;
}

void rcmd_destroy(rcmd_t *rcmd) {
}

/* ===================================== *
 * ==== tunnel socket establishment ==== *
 * ===================================== */

 //
 // modifies "opt_port" as a side-effect
 //

int create_listener(u_short port) {
struct sockaddr_in s_in;
socklen_t len;
char *range;
int i,s;

	debug("create_listener(port=%d)\n",port);
	if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
		die("socket");
	i=1;
	if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&i,sizeof(i))==-1)
		die("setsockopt");
	s_in.sin_family=AF_INET;
	s_in.sin_addr.s_addr=htonl(INADDR_ANY);
	/* let system choose port */
	range=getenv("GLOBUS_TCP_PORT_RANGE");
	if ((range=getenv("GLOBUS_TCP_PORT_RANGE"))==NULL || port) {
		s_in.sin_port=htons(port);
		if (bind(s,(struct sockaddr *)&s_in,sizeof(s_in))==-1)
			die("bind");
	}
	/* if we got a port-range, randomly
	 * select a port in between the range
	 *
	 * XXX use globus port allocation table,
	 * XXX if possible.
	 */
	else {
	uint16_t sh;
	int ntries;
	char *cp;
	int a,b;

		a=atoi(range);
		if ((cp=index(range,' '))==NULL &&
			(cp=index(range,':'))==NULL &&
			(cp=index(range,','))==NULL &&
			(cp=index(range,'-'))==NULL);
		b=cp?atoi(cp+1):65535;
		debug("port range %d %d\n",a,b);
		ntries=0;
		for(;;) {
			if (read_rand(&sh,2)!=2)
				die("read rand");
			sh=(uint16_t)(a+(b-a+1)*((float)sh/65536.0));
			s_in.sin_port=htons(sh);
			debug("sh=%d\n",sh);
			if (bind(s,(struct sockaddr *)&s_in,sizeof(s_in))==-1) {
				if (ntries==30)
					die("could not bind port range");
				usleep(10000);
				ntries++;
			}
			else {
				opt_port=sh;
				break;
			}
		}
	}
	if (listen(s,1)==-1)
		die("listen");
	len=sizeof(s_in);
	if (getsockname(s,(struct sockaddr *)&s_in,&len)==-1)
		die("getsockname");
	opt_port=ntohs(s_in.sin_port);
	debug("sockname port=%d\n",opt_port);
	return s;
}

int connect_nb(struct in_addr addr,u_short port) {
struct sockaddr_in s_in;
int s;

	debug("> connect_to(addr=%s,port=%hd)\n",
		inet_ntoa(addr),port);
	if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
		die("socket");
	//if (fcntl(s,F_SETFL,O_RDWR|O_NONBLOCK)==-1)
	if (fcntl(s,F_SETFL,O_RDWR|O_NONBLOCK)==-1)
		die("fcntl");
	s_in.sin_family=AF_INET;
	s_in.sin_addr=addr;
	s_in.sin_port=htons(port);
	if (connect(s,(struct sockaddr *)&s_in,sizeof(s_in))==-1
		&& errno!=EINPROGRESS) {
			// die("connect");
			close(s);
			return -1;
		}
	debug("connect_to: got socket=%d\n",s);
	return s;
}

/*
 * accept_once:
 *
 * accept a connection from a listener
 * and close the listener afterwards
 */

int accept_once(int sl) {
struct sockaddr_in s_in;
socklen_t len;
int s;

	len=sizeof(s_in);
	if ((s=accept(sl,(struct sockaddr *)&s_in,&len))==-1)
		die("accept");
	close(sl);
	return s;
}

/*
 * maybe_accept:
 *
 * accept a connection only
 * when the listening socket
 * has one
 */

#ifdef I2G
int maybe_accept(int s) {
struct timeval tv;
fd_set fds;

        FD_ZERO(&fds);
        FD_SET(s,&fds);
        tv.tv_sec=0;
        tv.tv_usec=0;
        if (select(s+1,&fds,NULL,NULL,&tv)==-1)
                die("select");
        return FD_ISSET(s,&fds)?accept_once(s):-1;
}
#endif

/*
 * sc==-1 means that in sl, an already existing
 *        connection is present.
 *
 * don't bother closing sockets - they will be
 * closed in the mainprogram.
 */

#ifdef I2G
int tun_establish(int sc,int sl) {
extern struct in_addr r_addr;	// used only as a hint
struct timeval tv;
fd_set rfds,wfds;
int s=-1,n;

	if (sc==-1)
		s=sl;
	else {
		n=sc>sl?sc:sl;
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(sl,&rfds);
		FD_SET(sc,&wfds);
		memset(&tv,0,sizeof(tv));
		if (do_remote)
			tv.tv_sec=5*60;
		else
			tv.tv_sec=3;
		if ((s=select(n+1,&rfds,&wfds,NULL,&tv))==-1)
			die("select");
		if (s==0)
			die("timeout");
#ifdef DEBUG
		if (do_remote==0) {
			fprintf(stderr,"nselect=%d\n",s);
			if (FD_ISSET(sl,&rfds))
				fprintf(stderr,"sl is set\n");
			if (FD_ISSET(sc,&wfds))
				fprintf(stderr,"sc is set\n");
		}
#endif
		if (s==0 && is_private(r_addr)) {
			fprintf(stderr,"warning: private target IP %s\n",inet_ntoa(r_addr));
			errno=ETIMEDOUT;
			die("select");
		}

#ifdef DEBUG
#if 1
		if (FD_ISSET(sl,&rfds))
			debug("sl avail, sl=%d\n",sl);
		if (FD_ISSET(sc,&wfds))
			debug("sc avail, sc=%d\n",sc);
#endif
#endif

		if (FD_ISSET(sl,&rfds)) {
			sl=accept_once(sl);
			debug("accepted sl=%d\n",sl);
			sc=-1;
			s=sl;
		}
		else if (FD_ISSET(sc,&wfds)) {
			if (fcntl(sc,F_SETFL,O_RDWR)==-1)
				die("fcntl");
			sl=-1;	// disable listener, else it might interfere
			s=sc;
		}
	}

	debug("sl=%d sc=%d\n",sl,sc);

	/* local instance sends startup signature */
	if (do_remote==0) {
		debug("sending startup on socket %d\n",s);
		n=0xfefefefe;
		if (write(s,&n,4)!=4)
			die("write");
		// note: we have to send a startup sequence.
		// closing the other socket is not enough.
		// reason: (1) it is not there (socketfd==-1) or
		//         (2) it is firewalled on the other end,
		//             a close() will not be noticed.
	}
	/* remote instance listen for startup */
	else {
		if (fcntl(sc,F_SETFL,O_RDWR)==-1)	/* unset NB */
			die("fcntl");
		FD_ZERO(&rfds);
		if (sc!=-1)
			FD_SET(sc,&rfds);
		if (sl!=-1)
			FD_SET(sl,&rfds);
		n=sc>sl?sc:sl;
		//tv.tv_sec=0;
		//tv.tv_usec=100000;
		//if (select(n+1,&rfds,NULL,NULL,NULL)==-1)
		tv.tv_sec=5*60;
		if (select(n+1,&rfds,NULL,NULL,&tv)<1)
			die("select");
		if (sc!=-1 && FD_ISSET(sc,&rfds))
			debug("startup: sc avail\n");
		if (sl!=-1 && FD_ISSET(sl,&rfds))
			debug("startup: sl avail\n");
		if (sc!=-1 && FD_ISSET(sc,&rfds))
			s=sc;
		else
			s=sl;
		//else die("BUG: no socket");

		if (read(s,&n,4)!=4)
			die("read");
		if (n!=0xfefefefe)
			die("strange tunnel handshake");
	}
	debug("establish: got socket=%d\n",s);
	return s;
}
#endif
