/*
 * Thu Jul  6 11:34:07 CET 2004 herp - Herbert Rosmanith
 *
 */

#include	"glogin.h"

#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/wait.h>
#include	<sys/utsname.h>
#include	<sys/socket.h>
#include	<sys/un.h>
#include	<syslog.h>
#include	<signal.h>
#include	<strings.h>
#include	<netinet/in.h>
#ifdef	DEBUG
#include	<arpa/inet.h>
#endif
#include	<netdb.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<errno.h>

#include	"resolv.h"
#include	"x11.h"

#include	"glogin.h"

extern struct utsname uts;
extern struct hostent *self_hp;

extern int do_remote;

u_short xdispno;	/* dispno */
char *xdisploc=NULL;	/* host:dispno */
char *xunixloc=NULL;	/* same for pf_unix */
char xcookie[16];

/*
 * ...sigh...
 *
 * should there be any error, we have to stuff it into
 * a pipe or else ... it can happen that it is sent to
 * the pty and result in unwanted commands there!
 */

void x11_exec(char *path,...) {
extern char **environ;
va_list ap;
char **args;
char msg[1024];
pid_t pid;
int p[2]={-1,-1};
int i,n;

	debug("> x11_exec path=%s\n",path);
	va_start(ap,path);
	n=0;
	for(n=0;va_arg(ap,char *);n++)
		;
	va_end(ap);
	if ((args=(char **)alloca((n+2)*sizeof(char *)))==NULL)
		die("out of memory");
	args[0]=basename(path);
	va_start(ap,path);
	for(i=1;i<=n;i++)
		args[i]=va_arg(ap,char *);
	args[i]=NULL;
	if (do_remote)
		if (pipe(p)==-1)
			return;
	switch(pid=fork()) {
	case -1:
		return;
	case 0:
		if (do_remote) {
			dup2(p[1],2);
			close(p[0]);
		}
		execve(path,args,environ);
		die("execve");
	default:
		break;
	}
	if (do_remote) {
		close(p[1]);
		while((i=read(p[0],msg,sizeof(msg)))>=1) {
			msg[i]=0;
			send_error_msg(0,msg);
		}
	}
	waitpid(pid,&i,0);
	if (do_remote)
		close(p[0]);
}

void x11_add_Xauthority(char *dsploc,char *xcookie_str) {
	debug("> x11_add_Xauthority dsploc=%s xcookie_str=%s\n",
		dsploc,xcookie_str);
	x11_exec(PATH_XAUTH,"-i","add",dsploc,"MIT-MAGIC-COOKIE-1",xcookie_str,NULL);
}

void x11_remove_Xauthority(char *dsploc) {
	debug("> x11_remove_Xauthority dsploc=%s\n",
		dsploc);
	//x11_exec(PATH_XAUTH,"-q","remove",dsploc,NULL);
	x11_exec(PATH_XAUTH,"-q","-i","remove",dsploc,NULL);
}

int x11_extract_cookie(char *display) {
u_char ch,buf[256];
struct sigaction sa,osa;
pid_t pid;
int p[2]={-1,-1};
int i;

        debug("> x11_extract_cookie display=%s\n",display);
        sa.sa_handler=SIG_IGN;
        sa.sa_flags=0;
        sigemptyset(&sa.sa_mask);
        sigaction(SIGCHLD,&sa,&osa);
        if (pipe(p)==-1)
                goto error_return;
        switch(pid=fork()) {
        case -1:
                die("fork");
        case 0:
                dup2(p[1],1);
                close(p[0]);
                // dup2(open("/dev/null",O_RDWR),2);
                execl(PATH_XAUTH,"xauth","extract","-",display,NULL);
                die("execl");
        default:
                break;
        }
        close(p[1]);
        waitpid(pid,&i,0);
        sigaction(SIGCHLD,&osa,NULL);
        if (read(p[0],&ch,1)!=1)        /* type 0=name 1=ipaddr */
                goto error_return;
        for(i=0;i<5;i++) {
                if (read(p[0],&ch,1)!=1)
                        goto error_return;
                if (read(p[0],buf,ch)!=ch)
                        goto error_return;
                buf[ch]=0;
                if (i<4 && read(p[0],buf,1)!=1)
                        goto error_return;
        }
        memcpy(xcookie,buf,16);
        return 0;
error_return:
	if (p[0]!=-1)
		close(p[0]);
	return -1;
}

int x11_create_listener() {
struct sockaddr_in saddr;
u_short port;
int i,s;
        /* x11-forwarder listener */
        if ((s=socket(PF_INET,SOCK_STREAM,0))==-1)
                die("socket");
        saddr.sin_family=AF_INET;
        saddr.sin_addr.s_addr=INADDR_ANY;	// XXX +x +X
        i=1;
        if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&i,sizeof(i))==-1)
                die("setsocktopt");
        for(port=6010;port<6100;port++) {
                saddr.sin_port=htons(port);
                if (bind(s,(struct sockaddr *)&saddr,sizeof(saddr))==0)        /* got it */
                        break;
        }
        if (port==6100)
                die("no free x11 forwarding ports");
        xdispno=port-6000;
        if (listen(s,16)==-1)
                die("listen");
        return s;
}

/* client: connecting to X-server */

static int disp_type=-1; /* 0-unix 1-inet */
static struct sockaddr_in saddr_disp;
static struct sockaddr_un uaddr_disp={AF_UNIX,"/tmp/.X11-unix/X*xxx"};

int x11_init_display() {
char *display;
char *cp;
int i;

        cp=index(uaddr_disp.sun_path,'*');
        disp_type=0;
        if ((display=getenv("DISPLAY"))==NULL) {
                // strcpy(cp,"0");
                return -1;
        }
        if (display[0]==':') {
                sprintf(cp,"%d",atoi(display+1));
                return 0;
        }
        disp_type=1;
        saddr_disp.sin_family=AF_INET;
        if ((cp=index(display,':'))==NULL)
		return -1;
        *cp=0;
        i=resolv(display,&saddr_disp.sin_addr);
        *cp++=':';	/* don't destroy DISPLAY= */
        if (i==-1)
		return -1;
        saddr_disp.sin_port=htons(6000+atoi(cp));
	return 0;
}

int x11_open_display() {
int s=-1;
	debug("> x11_open_display display_type=%d\n",disp_type);
	if (disp_type==0) {
		if ((s=socket(PF_UNIX,SOCK_STREAM,0))==-1)
			goto error_return;
		debug("  created unix socket s=%d\n",s);
		if (connect(s,(struct sockaddr *)&uaddr_disp,sizeof(uaddr_disp))==-1)
			goto error_return;
		return s;
	}
	if (disp_type==1) {
		if ((s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP))==-1)
			goto error_return;
		debug("  created inet socket s=%d\n",s);
		debug("  connect: addr=%s\n",inet_ntoa(saddr_disp.sin_addr));
		debug("  connect: port=%d\n",ntohs(saddr_disp.sin_port));
		if (connect(s,(struct sockaddr *)&saddr_disp,sizeof(saddr_disp))==-1)
			goto error_return;
		debug("  connect: s=%d\n",s);
		return s;
	}
	/* fallthrough */
error_return:
	if (s!=-1)
		close(s);
	return -1;
}

/*
 * which cookie to use depends on whether or not the
 * display is on the same machine we started glogin
 * if it is not, we have to use the cookie of the
 * remote X-server, since glogin doesnt alter the
 * contents of the X11-datastream (unlike ssh).
 *
 * this function is called by local glogin instance only.
 *
 */
int x11_init_cookie() {
char *display,*dispnode,*cp,*cp2;
char *dfqdn;
struct in_addr addr;
int i;

	debug("> x11_init_cookie\n");

	if ((display=getenv("DISPLAY"))==NULL)
		goto local_display_return;
	if ((cp=index(display,':'))==NULL)
		return -1;
	if ((cp2=index(cp,'.')))	/* zap screen# */
		*cp2=0;
	if (cp==display)
		goto local_display_return;
	if ((cp2=index(display,'/'))) {
		if (strncmp(cp2+1,"unix",4)==0 && cp2+5==cp)
			goto local_display_return;
	}
	// fprintf(stderr,"display=%s dispno=%d\n",
	// 	display,atoi(cp+1));
	if (atoi(cp+1)>=10)	/* bah! sshd doesn't loose root! */
		goto remote_display_return;
	*cp=0;
	dispnode=(char *)alloca(strlen(display)+1);
	strcpy(dispnode,display);
	*cp=':';

	if (resolv(dispnode,&addr)==-1)
		return -1;
	if (ntohl(addr.s_addr)==INADDR_LOOPBACK)	/* localhost? */
		goto local_display_return;
	// fprintf(stderr,"uts.nodename=%s dispnode=%s\n",uts.nodename,dispnode);
	if (strcmp(uts.nodename,dispnode)==0)
		goto local_display_return;
	dfqdn=dnslookup(dispnode);
	// fprintf(stderr,"dfqdn=%s self_hp->h_name=%s\n",dfqdn,self_hp->h_name);
	i=(strcmp(dfqdn,self_hp->h_name)==0);
	free(dfqdn);
	if (i)
		goto local_display_return;
	/* okay, so we now assume we have a remote DISPLAY
	 * try to get a cookie for that DISPLAY first.
	 * if that fails, proceed in the same way as with
	 * a local DISPLAY
	 */

remote_display_return:
	//fprintf(stderr,"found remote display\n");
	// printf("extract remote display=%s\n",display);
	if (strncmp(display,"localhost:",10)==0) {
		cp=(char *)alloca(strlen(uts.nodename)+5+6+1);
		sprintf(cp,"%s/unix%s",uts.nodename,display+9);
		display=cp;
	}
	// fprintf(stderr,"extract remote display=%s\n",display);
	// syslog(LOG_NOTICE|LOG_DAEMON,"REMOTE DISPLAY\n");
	if (x11_extract_cookie(display)==0)
		return 0;
	/* fallthru */

local_display_return:

	// XXX - PROBLEMS WITH X11-COOKIES IN CONJUNCTION WITH 
	// KDM! (xserver has different auth-file ?!(/")(/&$(

	goto remote_display_return;

	// fprintf(stderr,"found local display\n");
	debug("Generating random cookie\n");
	return read_rand(xcookie,16);
}

